module vars
  implicit none
  public :: phi,apar,bpar,jpar
  public :: xsheet,ysheet
  private

  complex, allocatable :: phi(:,:),apar(:,:),bpar(:,:),jpar(:,:)
  real :: xsheet=.5, ysheet=.5

end module vars

module funcs
  implicit none
  public :: jfunc, djfunc
  private

contains

  function jfunc(x)
    use vars, only: jpar, ysheet
    use constants, only: zi,pi
    use kgrids, only: akx,aky,nakx,naky,y0
    real :: jfunc
    real, intent(in) :: x
    integer :: it,ik
    jfunc=0.
    do it=1,nakx
       jfunc=jfunc &
            & + real(2.*jpar(it,1)*exp(zi*(akx(it)*x+aky(1)*ysheet))) &
            & + sum(real(jpar(it,2:naky) * &
            & exp(zi*(akx(it)*x+aky(2:naky)*ysheet))))
    end do
  end function jfunc

  function djfunc(x)
    use vars, only: jpar, ysheet
    use constants, only: zi,pi
    use kgrids, only: akx,aky,nakx,naky,y0
    real :: djfunc
    real, intent(in) :: x
    integer :: it,ik
    djfunc=0.
    do it=1,nakx
       djfunc=djfunc &
            & + real(2.*zi*akx(it)*jpar(it,1) &
            & * exp(zi*(akx(it)*x+aky(1)*ysheet))) &
            & + sum(real(zi*akx(it)*jpar(it,2:naky) &
            & * exp(zi*(akx(it)*x+aky(2:naky)*ysheet))))
    end do
  end function djfunc

end module funcs

program tearing_diag
  use constants, only: pi
  use netcdf, only: NF90_NOWRITE, NF90_NOERR
  use netcdf, only: nf90_open
  use netcdf, only: nf90_inquire_dimension, nf90_inq_dimid
  use netcdf, only: nf90_inq_varid, nf90_get_var
  use netcdf_utils, only: kind_nf, netcdf_error
  use mp, only: init_mp
  use file_utils, only: init_file_utils, run_name, stdout=>stdout_unit
  use file_utils, only: get_unused_unit
  use run_parameters, only: init_run_parameters, beta
  use kgrids, only: init_kgrids, nx, ny, kperp2
  use convert, only: r2c
  use agk_transforms, only: inverse2, transform2
  use dist_fn, only: init_dist_fn
  use agk_mem, only: init_mem

  use vars, only: phi,apar,bpar,jpar
  use vars, only: xsheet, ysheet
  use funcs, only: jfunc, djfunc

  implicit none

  integer (kind=kind_nf) :: file_id
  integer (kind=kind_nf) :: kxdimid,kydimid
  integer (kind=kind_nf) :: kxvarid,kyvarid
  integer (kind=kind_nf) :: time_id,tdim_id
  integer (kind=kind_nf) :: phi_id,apar_id,bpar_id

  integer :: nkx,nky,nnx,nny
  integer :: itime,ntime
  integer :: i,j
  integer :: unit
  integer :: istatus
  integer :: xcut1,xcut2,xcut3,xcut4,ycut1,ycut2,ycut3,ycut4

  real, parameter :: del=1.e-3
  real :: width,width_guess,jmax_guess,xjmax,width2
  real :: time
  real :: dkx,dky,lx,ly,x,y
  real, allocatable :: phi_ri(:,:,:),apar_ri(:,:,:),bpar_ri(:,:,:)
  real, allocatable :: phi_xy(:,:), phi_eq_xy(:,:)
  real, allocatable :: apar_xy(:,:),apar_eq_xy(:,:)
  real, allocatable :: bpar_xy(:,:),bpar_eq_xy(:,:)
  real, allocatable :: jpar_xy(:,:),jpar_eq_xy(:,:)
  real :: growth, time_old, apar_old
  character (len=300) :: filename
  character (len=300) :: eqfile='equilibrium.dat'
  logical :: list, ex
  logical :: debug=.false.

  call init_mp
  call init_file_utils(list)
  call init_mem
  call init_run_parameters
  call init_dist_fn
  call init_kgrids

  xcut1=1
  xcut2=nx/4+1
  xcut3=nx/2+1
  xcut4=nx/2+nx/4+1
  ycut1=1
  ycut2=ny/4+1
  ycut3=ny/2+1
  ycut4=ny/2+ny/4+1

  filename=trim(trim(run_name)//'.out.nc')
  istatus=nf90_open(filename, NF90_NOWRITE, file_id)
  if (istatus /= NF90_NOERR) call netcdf_error (istatus, file_id)
  if(debug) write(stdout,*) 'DEBUG:: input file ', &
       & 'filename= ', trim(filename), &
       & ' file_id= ', file_id

  istatus = nf90_inq_dimid (file_id, "kx", kxdimid)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, dim='kx')
  istatus = nf90_inq_dimid (file_id, "ky", kydimid)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, dim='ky')
  istatus = nf90_inquire_dimension (file_id, kxdimid, len=nkx)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, dimid=kxdimid)
  istatus = nf90_inquire_dimension (file_id, kydimid, len=nky)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, dimid=kydimid)

  istatus = nf90_inq_varid (file_id, "kx", kxvarid)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, var='kx')
  istatus = nf90_inq_varid (file_id, "ky", kyvarid)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, var='ky')
  istatus = nf90_get_var (file_id, kxvarid, dkx, start=(/ 2 /))
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, kxvarid)
  istatus = nf90_get_var (file_id, kyvarid, dky, start=(/ 2 /))
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, kyvarid)
  lx=2.*pi/dkx; ly=2.*pi/dky
  xsheet=xsheet*lx
  ysheet=ysheet*ly
  if(debug) write(stdout,*) 'DEBUG:: inquire grid ', &
       & 'kxdimid= ', kxdimid, &
       & ' kydimid= ', kydimid, &
       & ' nkx =     ', nkx, &
       & ' nky =     ', nky, &
       & ' dkx =     ', dkx, &
       & ' dky =     ', dky, &
       & ' lx =     ', lx, &
       & ' ly =     ', ly

  istatus = nf90_inq_varid (file_id, "t", time_id)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, var='t')

  istatus = nf90_inq_dimid (file_id, "t", tdim_id)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, var='t')
  istatus = nf90_inquire_dimension (file_id, tdim_id, len=ntime)
  if (istatus /= NF90_NOERR) &
       & call netcdf_error (istatus, file_id, dimid=tdim_id)
  if(debug) write(stdout,*) 'DEBUG:: inquire time ', &
       & 'time_id', time_id, &
       & ' ntime', ntime

  istatus = nf90_inq_varid (file_id, 'phi0', phi_id)
  if (istatus /= NF90_NOERR) call netcdf_error &
       & (istatus, file_id, var='phi0')
  if(debug) write(stdout,*) 'DEBUG:: inquire variable ', &
       & 'phi_id', phi_id
  istatus = nf90_inq_varid (file_id, 'apar0', apar_id)
  if (istatus /= NF90_NOERR) call netcdf_error &
       & (istatus, file_id, var='apar0')
  if(debug) write(stdout,*) 'DEBUG:: inquire variable ', &
       & 'apar_id', apar_id
  istatus = nf90_inq_varid (file_id, 'bpar0', bpar_id)
  if (istatus /= NF90_NOERR) call netcdf_error &
       & (istatus, file_id, var='bpar0')
  if(debug) write(stdout,*) 'DEBUG:: inquire variable ', &
       & 'bpar_id', bpar_id

  if(nx>nkx) then
     nnx=nx
  else
     nnx=(3*nkx+1)/2
  endif
  if(ny>nky) then
     nny=ny
  else
     nny=3*nky
  endif

  allocate(phi_xy(nx,ny),apar_xy(nx,ny),bpar_xy(nx,ny),jpar_xy(nx,ny))
  phi_xy(:,:)=0.; apar_xy(:,:)=0.; bpar_xy(:,:)=0.; jpar_xy(:,:)=0.

  allocate(phi_eq_xy(nx,ny), apar_eq_xy(nx,ny))
  allocate(bpar_eq_xy(nx,ny),jpar_eq_xy(nx,ny))
  phi_eq_xy(:,:)=0.; apar_eq_xy(:,:)=0.
  bpar_eq_xy(:,:)=0.;jpar_eq_xy(:,:)=0.

  inquire(file=eqfile,exist=ex)
  if(ex) then
     call get_unused_unit(unit)
     open(unit,file=eqfile,form='unformatted')
     read(unit) phi_eq_xy,apar_eq_xy,bpar_eq_xy,jpar_eq_xy
     close(unit)
  else
     stop 'equilibrium file does not exists'
  endif

  allocate(phi_ri(1:2,1:nkx,1:nky), phi(1:nkx,1:nky))
  allocate(apar_ri(1:2,1:nkx,1:nky),apar(1:nkx,1:nky))
  allocate(bpar_ri(1:2,1:nkx,1:nky),bpar(1:nkx,1:nky))
  allocate(jpar(1:nkx,1:nky))

  phi_ri=0.; apar_ri=0.; bpar_ri=0.
  phi=cmplx(0.,0.); apar=cmplx(0.,0.); bpar=cmplx(0.,0.)
  jpar=cmplx(0.,0.)

  open(unit=51,file='xprof1.dat',status='unknown')
  open(unit=52,file='xprof2.dat',status='unknown')
  open(unit=53,file='xprof3.dat',status='unknown')
  open(unit=54,file='xprof4.dat',status='unknown')
  open(unit=55,file='yprof1.dat',status='unknown')
  open(unit=56,file='yprof2.dat',status='unknown')
  open(unit=57,file='yprof3.dat',status='unknown')
  open(unit=58,file='yprof4.dat',status='unknown')
  open(unit=61,file='growth_rate.dat',status='unknown')
  open(unit=71,file='current_width.dat',status='unknown')

  if(debug) open(91,file='function.dat',status='unknown')

  write(51,'(a,f12.8)') '# x profile at y = ',ly/ny*(ycut1-1)
  write(52,'(a,f12.8)') '# x profile at y = ',ly/ny*(ycut2-1)
  write(53,'(a,f12.8)') '# x profile at y = ',ly/ny*(ycut3-1)
  write(54,'(a,f12.8)') '# x profile at y = ',ly/ny*(ycut4-1)
  write(55,'(a,f12.8)') '# y profile at x = ',lx/nx*(xcut1-1)
  write(56,'(a,f12.8)') '# y profile at x = ',lx/nx*(xcut2-1)
  write(57,'(a,f12.8)') '# y profile at x = ',lx/nx*(xcut3-1)
  write(58,'(a,f12.8)') '# y profile at x = ',lx/nx*(xcut4-1)
  write(51,'(a1,a19,5(1x,a19))') '#','time','x', &
       'phi','apar','bpar','current'
  write(52,'(a1,a19,5(1x,a19))') '#','time','x', &
       'phi','apar','bpar','current'
  write(53,'(a1,a19,5(1x,a19))') '#','time','x', &
       'phi','apar','bpar','current'
  write(54,'(a1,a19,5(1x,a19))') '#','time','x', &
       'phi','apar','bpar','current'
  write(55,'(a1,a19,5(1x,a19))') '#','time','y', &
       'phi','apar','bpar','current'
  write(56,'(a1,a19,5(1x,a19))') '#','time','y', &
       'phi','apar','bpar','current'
  write(57,'(a1,a19,5(1x,a19))') '#','time','y', &
       'phi','apar','bpar','current'
  write(58,'(a1,a19,5(1x,a19))') '#','time','y', &
       'phi','apar','bpar','current'

  timeloop : do itime=1,ntime
     istatus = nf90_get_var (file_id, time_id, time, start=(/ itime /))
     if (istatus /= NF90_NOERR) &
          & call netcdf_error (istatus, file_id, varid=time_id)
     istatus = nf90_get_var (file_id, phi_id, phi_ri(1:2,1:nkx,1:nky), &
          & start=(/ 1,1,1,itime /),  count=(/ 2,nkx,nky,1 /))
     if (istatus /= NF90_NOERR) &
          & call netcdf_error (istatus, file_id, phi_id)
     call r2c(phi,phi_ri)
     istatus = nf90_get_var (file_id, apar_id, apar_ri(1:2,1:nkx,1:nky), &
          & start=(/ 1,1,1,itime /),  count=(/ 2,nkx,nky,1 /))
     if (istatus /= NF90_NOERR) &
          & call netcdf_error (istatus, file_id, apar_id)
     call r2c(apar,apar_ri)
     istatus = nf90_get_var (file_id, bpar_id, bpar_ri(1:2,1:nkx,1:nky), &
          & start=(/ 1,1,1,itime /),  count=(/ 2,nkx,nky,1 /))
     if (istatus /= NF90_NOERR) &
          & call netcdf_error (istatus, file_id, bpar_id)
     call r2c(bpar,bpar_ri)

     jpar(:,:)=apar(:,:)*kperp2(:,:)/beta*.5

     call transform2(phi, phi_xy,nny,nnx)
     call transform2(apar,apar_xy,nny,nnx)
     call transform2(bpar,bpar_xy,nny,nnx)
     call transform2(jpar,jpar_xy,nny,nnx)

     phi_xy =phi_xy -phi_eq_xy
     apar_xy=apar_xy-apar_eq_xy
     bpar_xy=bpar_xy-bpar_eq_xy
     jpar_xy=jpar_xy-jpar_eq_xy

     call inverse2(jpar_xy,jpar,nny,nnx)

     do i=1,nx
        x=lx/nx*(i-1)
        write(51,'(6d20.12)') time,x, &
             & phi_xy(i,ycut1),apar_xy(i,ycut1), &
             & bpar_xy(i,ycut1),jpar_xy(i,ycut1)
        write(52,'(6d20.12)') time,x, &
             & phi_xy(i,ycut2),apar_xy(i,ycut2), &
             & bpar_xy(i,ycut2),jpar_xy(i,ycut2)
        write(53,'(6d20.12)') time,x, &
             & phi_xy(i,ycut3),apar_xy(i,ycut3), &
             & bpar_xy(i,ycut3),jpar_xy(i,ycut3)
        write(54,'(6d20.12)') time,x, &
             & phi_xy(i,ycut4),apar_xy(i,ycut4), &
             & bpar_xy(i,ycut4),jpar_xy(i,ycut4)
     end do
     write(51,*);write(52,*);write(53,*);write(54,*)
     do j=1,ny
        y=ly/ny*(j-1)
        write(55,'(6d20.12)') time,y, &
             & phi_xy(xcut1,j),apar_xy(xcut1,j), &
             & bpar_xy(xcut1,j),jpar_xy(xcut1,j)
        write(56,'(6d20.12)') time,y, &
             & phi_xy(xcut2,j),apar_xy(xcut2,j), &
             & bpar_xy(xcut2,j),jpar_xy(xcut2,j)
        write(57,'(6d20.12)') time,y, &
             & phi_xy(xcut3,j),apar_xy(xcut3,j), &
             & bpar_xy(xcut3,j),jpar_xy(xcut3,j)
        write(58,'(6d20.12)') time,y, &
             & phi_xy(xcut4,j),apar_xy(xcut4,j), &
             & bpar_xy(xcut4,j),jpar_xy(xcut4,j)
     end do
     write(55,*);write(56,*);write(57,*);write(58,*)

     if(itime/=1) then
        growth=(apar_xy(xcut3,ycut3)-apar_old)/(time-time_old)/apar_xy(xcut3,ycut3)
        write(61,'(3d20.12)') time,apar_xy(xcut3,ycut3),growth
     endif
     time_old=time; apar_old=apar_xy(xcut3,ycut3)

     if(debug) then
        do i=1,nx
           x=lx/nx*(i-1)
           write(91,'(5d20.12)') time,x,jpar_xy(i,ycut3),jfunc(x),djfunc(x)
        end do
        write(91,*)
     endif

     if(itime/=1) then
        width_guess=xsheet
        guess1 : do while (jfunc(xsheet)*jfunc(width_guess)>=0.)
           width_guess=width_guess+del*lx
           if(width_guess>=lx) then
              width_guess=xsheet
              exit guess1
           endif
        end do guess1
        jmax_guess=width_guess
        guess2 : do while (djfunc(width_guess)*djfunc(jmax_guess)>=0.)
           jmax_guess=jmax_guess+del*lx
           if(jmax_guess>=lx) then
              jmax_guess=width_guess
              exit guess2
           endif
        end do guess2
        ! find position where dj/dx=0
        call root_finder_bisect(width_guess,jmax_guess, &
             0.,xjmax,djfunc)

        width=xsheet; width2=xsheet
        call root_finder_bisect(xsheet,width_guess, &
             .5*jfunc(xsheet),width,jfunc)
        call root_finder_bisect(width,jmax_guess, &
             .5*(jfunc(xsheet)+jfunc(xjmax)),width2,jfunc)
        width=2.*(width-xsheet)
        width2=2.*(width2-xsheet)

        write(71,'(4d20.12)') time,jpar_xy(nx/2+1,ycut3),width,width2
     endif
  end do timeloop

  close(51); close(52); close(53); close(54)
  close(55); close(56); close(57); close(58)
  close(61)
  close(71)

  if(debug) close(91)

contains

  subroutine root_finder_bisect (g0, g1, val, sol, func)

    real, intent(in) :: g0, g1, val
    real, intent(out) :: sol

    interface
       function func (x)
         real, intent(in) :: x
         real :: func
       end function func
    end interface

    real :: x0, x1, xc, y0, y1, yc

    x0 = g0; x1 = g1
    ! rhs - lhs
    y0 = func(x0) - val
    y1 = func(x1) - val

    if (y0*y1 > 0.0) then
       write (0,*) 'no solution in specified domain'
       !       stop
       return
    end if

    do while (abs((x0-x1)/max(abs(x0),abs(x1))) > 1.e-10)

       xc = (x0+x1)/2.0
       yc = func(xc) - val

       if (y0*yc > 0.0) then
          x0 = xc; y0 = yc
       else
          x1 = xc; y1 = yc
       end if

    end do

    if (abs(y0) < abs(y1)) then
       sol = x0
    else
       sol = x1
    end if

  end subroutine root_finder_bisect

end program tearing_diag
