!=============================================================================!
!=============================================================================!
!!*                                                                         *!!
!!              AstroGK Memory Management and Diagnostic Module              !!
!!                                   2008                                    !!
!!*                                                                         *!!
!=============================================================================!
!=============================================================================!
! This module contains routines for the accounting of the allocation of large
!     arrays in AstroGK.  It prints out memory diagnostics to a file
!     runname.mem, with a running account of total memory allocated.
!NAMELIST memory:
!   mem_account     !T=turn on accounting statistics
!   mem_output_on   !T=Output running tally of memory
!   mem_debug       !T=Turn on output of allocated variables with dimensions
!NOTE: These routines are only run from proc0
module agk_mem
  use agk_time, only: time
  implicit none
  private
  !Variables specifying allocated variable type
  integer, parameter :: tcomplex = 0
  integer, parameter :: treal = 1
  integer, parameter :: tinteger = 2
  integer, parameter :: ttotal = 3
  character(8), dimension(0:3) :: ttype
  integer :: mem_unit                         !Unit for run_name.mem file
  logical :: mem_account           !T=turn on accounting statistics
  logical :: mem_output_on         !T=Output running tally of memory
  logical :: mem_debug             !T=Turn on output of allocated variables with dimensions
  logical :: initialized = .false.
  real, dimension(0:3) :: memsize        !Size allocated of each type(MB)
  real :: memmax                         !Maximum allocated memory (MB)
  real, dimension(0:2) :: tbytes         !Megabytes for each type (MB)
  logical :: no_driver = .false.

  interface attempt_alloc
     module procedure attempt_allocate1,attempt_allocate2,attempt_allocate3
     module procedure attempt_allocate4,attempt_allocate5,attempt_allocate6
  end interface

  interface mem_dealloc
     module procedure mem_deallocate1,mem_deallocate2,mem_deallocate3
     module procedure mem_deallocate4,mem_deallocate5,mem_deallocate6
  end interface

  public :: tcomplex,treal,tinteger
  public :: init_mem,finish_mem
  public :: attempt_alloc,mem_dealloc,complete_alloc
  public :: highwater_memory
  public :: memmax

contains
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! Open output file run_name.mem if mem_output_on=.true. and on proc0
  subroutine init_mem
    use file_utils, only: open_output_file
    use mp, only: proc0
    implicit none

    if (initialized) return
    initialized = .true.

    call read_parameters
    
    if (.not. proc0 .or. .not. mem_account) return

    if (mem_output_on .or. mem_debug) call open_output_file(mem_unit,".mem")

    !Initialize allocated memory of each type
    memsize(:)=0
    memmax=memsize(ttotal)
    !Set variable type names and sizes
    ttype(tcomplex)="complex"
    ttype(treal)="real"
    ttype(tinteger)="integer"
    ttype(ttotal)="total"
    !Set type sizes (Mbytes) 
    !NOTE: Assumes promotion to double precision during compilation
    tbytes(tcomplex)=16./1024.**2.
    tbytes(treal)=8./1024.**2.
    tbytes(tinteger)=8./1024.**2.
    !For output of memory accounting, temporarily set time=0
    time=0.

  end subroutine init_mem
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! Read parameters from memory namelist
  subroutine read_parameters 
    use file_utils
    use mp, only: proc0, broadcast
    implicit none
    integer :: in_file
    logical :: exist

    namelist /memory/ mem_account, mem_output_on, mem_debug
    
    !Default values
    mem_account=.true.
    mem_output_on=.false.
    mem_debug=.false.

    if (proc0) then
       in_file=input_unit_exist("memory",exist)
       if (.not. exist) then
          no_driver = .true.
       else
          read (unit=input_unit("memory"), nml=memory)
       endif
    endif

    call broadcast (mem_account)
    call broadcast (mem_output_on)
    call broadcast (mem_debug)
    call broadcast (no_driver)

  end subroutine read_parameters
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! Close output file run_name.mem if mem_output_on=.true. and on proc0
  subroutine finish_mem
    use file_utils, only: close_output_file
    use mp, only: proc0
    implicit none

    if (.not. proc0 .or. .not. mem_account) return

    if (mem_output_on .or. mem_debug) call close_output_file(mem_unit)

  end subroutine finish_mem
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 1-dimensional array
! Attempt to allocate variable varname. Account for size and output running tally
  subroutine attempt_allocate1(varname,ntype,ndim,n1)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1                     !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 1) write(mem_unit,'(a,a)') 'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,1(i12,a2),6x,f10.3)')'Allocating ',ttype(ntype),varname, &
         ' ( ',n1,' )',tbytes(ntype)*real(n1)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) + tbytes(ntype)*real(n1)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
    
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         tbytes(ntype)*real(n1),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)

    call flush_output_file(mem_unit,".mem")
        
  end subroutine attempt_allocate1
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 2-dimensional array
! Attempt to allocate variable varname. Account for size and output running tally
  subroutine attempt_allocate2(varname,ntype,ndim,n1,n2)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2                  !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 2) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,2(i12,a2),i12,6x,f10.3)')'Allocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,' )',n1*n2,tbytes(ntype)*real(n1*n2)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) + tbytes(ntype)*real(n1*n2)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
    
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         tbytes(ntype)*real(n1*n2),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine attempt_allocate2
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 3-dimensional array
! Attempt to allocate variable varname. Account for size and output running tally
  subroutine attempt_allocate3(varname,ntype,ndim,n1,n2,n3)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3               !Dimensions

    if (.not. proc0 .or. .not. mem_account) return
    
    !ERROR CHECK
    if (ndim .ne. 3) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,3(i12,a2),i12,6x,f10.3)')'Allocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,', ',n3,' )',n1*n2*n3,tbytes(ntype)*real(n1*n2*n3)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) + tbytes(ntype)*real(n1*n2*n3)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
    
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         tbytes(ntype)*real(n1*n2*n3),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine attempt_allocate3
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 4-dimensional array
! Attempt to allocate variable varname. Account for size and output running tally
  subroutine attempt_allocate4(varname,ntype,ndim,n1,n2,n3,n4)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3,n4            !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 4) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,4(i12,a2),i12,6x,f10.3)')'Allocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,', ',n3,', ',n4,' )',n1*n2*n3*n4,tbytes(ntype)*real(n1*n2*n3*n4)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) + tbytes(ntype)*real(n1*n2*n3*n4)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
    
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         tbytes(ntype)*real(n1*n2*n3*n4),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine attempt_allocate4
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 5-dimensional array
! Attempt to allocate variable varname. Account for size and output running tally
  subroutine attempt_allocate5(varname,ntype,ndim,n1,n2,n3,n4,n5)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3,n4,n5         !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 5) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,5(i12,a2),i12,6x,f10.3)')'Allocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,', ',n3,', ',n4,', ',n5,' )',n1*n2*n3*n4*n5,tbytes(ntype)*real(n1*n2*n3*n4*n5)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) + tbytes(ntype)*real(n1*n2*n3*n4*n5)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
    
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         tbytes(ntype)*real(n1*n2*n3*n4*n5),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine attempt_allocate5
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 6-dimensional array
! Attempt to allocate variable varname. Account for size and output running tally
  subroutine attempt_allocate6(varname,ntype,ndim,n1,n2,n3,n4,n5,n6)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3,n4,n5,n6      !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 6) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,6(i12,a2),i12,6x,f10.3)')'Allocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,', ',n3,', ',n4,', ',n5,', ',n6,' )',n1*n2*n3*n4*n5*n6,tbytes(ntype)*real(n1*n2*n3*n4*n5*n6)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) + tbytes(ntype)*real(n1*n2*n3*n4*n5*n6)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
    
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         tbytes(ntype)*real(n1*n2*n3*n4*n5*n6),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine attempt_allocate6
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 1-dimensional array
! Deallocate variable varname. Account for size and output running tally
  subroutine mem_deallocate1(varname,ntype,ndim,n1)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1                     !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 1) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,1(i12,a2),6x,f10.3)')'Deallocating ',ttype(ntype),varname, &
         ' ( ',n1,' )',tbytes(ntype)*real(n1)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) - tbytes(ntype)*real(n1)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
        
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         -tbytes(ntype)*real(n1),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine mem_deallocate1
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 2-dimensional array
! Deallocate variable varname. Account for size and output running tally
  subroutine mem_deallocate2(varname,ntype,ndim,n1,n2)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2                  !Dimensions

     if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 2) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,2(i12,a2),6x,f10.3)')'Deallocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,' )',tbytes(ntype)*real(n1*n2)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) - tbytes(ntype)*real(n1*n2)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
        
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         -tbytes(ntype)*real(n1*n2),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine mem_deallocate2
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 3-dimensional array
! Deallocate variable varname. Account for size and output running tally
  subroutine mem_deallocate3(varname,ntype,ndim,n1,n2,n3)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3               !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 3) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,3(i12,a2),6x,f10.3)')'Deallocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,', ',n3,' )',tbytes(ntype)*real(n1*n2*n3)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) - tbytes(ntype)*real(n1*n2*n3)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
        
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         -tbytes(ntype)*real(n1*n2*n3),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine mem_deallocate3
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 4-dimensional array
! Deallocate variable varname. Account for size and output running tally
  subroutine mem_deallocate4(varname,ntype,ndim,n1,n2,n3,n4)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3,n4            !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 4) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,4(i12,a2),6x,f10.3)')'Deallocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,', ',n3,', ',n4,' )', tbytes(ntype)*real(n1*n2*n3*n4)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) - tbytes(ntype)*real(n1*n2*n3*n4)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
    if (memsize(ttotal) .gt. memmax) memmax=memsize(ttotal)
        
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         -tbytes(ntype)*real(n1*n2*n3*n4),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine mem_deallocate4
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 5-dimensional array
! Deallocate variable varname. Account for size and output running tally
  subroutine mem_deallocate5(varname,ntype,ndim,n1,n2,n3,n4,n5)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3,n4,n5         !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 5) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,5(i12,a2),6x,f10.3)')'Deallocating ',ttype(ntype),varname, &
          ' ( ',n1,', ',n2,', ',n3,', ',n4,', ',n5,' )',tbytes(ntype)*real(n1*n2*n3*n4*n5)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) - tbytes(ntype)*real(n1*n2*n3*n4*n5)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
        
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         -tbytes(ntype)*real(n1*n2*n3*n4*n5),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine mem_deallocate5
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! 6-dimensional array
! Deallocate variable varname. Account for size and output running tally
  subroutine mem_deallocate6(varname,ntype,ndim,n1,n2,n3,n4,n5,n6)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate
    integer, intent(in) :: ntype                  !Variable type
    integer, intent(in) :: ndim                   !Number of dimensions
    integer, intent(in) :: n1,n2,n3,n4,n5,n6      !Dimensions

    if (.not. proc0 .or. .not. mem_account) return

    !ERROR CHECK
    if (ndim .ne. 6) write(mem_unit,'(a,a)')'ERROR: Number of dimensions incorrect for ',varname

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a,a3,6(i12,a2),6x,f10.3)')'Deallocating ',ttype(ntype),varname, &
         ' ( ',n1,', ',n2,', ',n3,', ',n4,', ',n5,', ',n6,' )',tbytes(ntype)*real(n1*n2*n3*n4*n5*n6)
        
    !Account for size of allocation
    memsize(ntype)=memsize(ntype) - tbytes(ntype)*real(n1*n2*n3*n4*n5*n6)
    !Sum for totals.
    memsize(ttotal)=sum(memsize(0:2))
        
    !Write accounting of total sizes
    if (mem_output_on) write(mem_unit,'(es14.6,f10.3,1x,a8,1x,a16,1x,4(a8,f10.3,1x))')time, &
         -tbytes(ntype)*real(n1*n2*n3*n4*n5*n6),ttype(ntype),varname, &
         ttype(0),memsize(0),ttype(1),memsize(1),ttype(2),memsize(2),ttype(3),memsize(3)
        
    call flush_output_file(mem_unit,".mem")
        
  end subroutine mem_deallocate6
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! Signal completed allocation of variable varname.
  subroutine complete_alloc(varname)
    use mp, only: proc0
    use file_utils, only: flush_output_file
    implicit none
    character (*), intent(in) :: varname          !Name of variable to allocate

    if (.not. proc0 .or. .not. mem_account) return

    !Output line about attempted allocation
    if (mem_debug) write(mem_unit,'(a,a8,a)')'Allocation of  ',varname,' complete'

    call flush_output_file(mem_unit,".mem")
        
  end subroutine complete_alloc
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! Output High-water memory at end of run
  subroutine highwater_memory
    use mp, only: proc0
    implicit none

    if (.not. proc0 .or. .not. mem_account) return

    PRINT '(/,'' High-Water Memory'',T25,0pf12.3,'' MB'',/)',memmax
        
  end subroutine highwater_memory
!------------------------------------------------------------------------------
!------------------------------------------------------------------------------
end module agk_mem


!Dependencies
! agk_mem.o: mp.o file_utils.o
!NOTES:
!1. mem_output_on=.true. needed to output run_name.mem file
