module collisions

  use redistribute, only: redist_type

  implicit none

  public :: init_collisions
  public :: solfp1
  public :: reset_init
  public :: update_vnewh          !Used for adapative hypercollisionality

  private

  ! knobsls

  logical :: conserve_momentum, const_v
  logical :: conserve_moments          ! should replace conserve_momentum -- MAB
  integer :: collision_model_switch
  logical :: adjust
  logical :: heating
  logical :: hyper_colls
  logical :: test

  integer, parameter :: collision_model_lorentz = 1
  integer, parameter :: collision_model_none = 3
  integer, parameter :: collision_model_lorentz_test = 5
! MAB>
  integer, parameter :: collision_model_full = 6
  integer, parameter :: collision_model_ediffuse = 7
! <MAB

! TT> First index omitted.
! GS2 has tunits(ik) factor which is removed in AstroGK.
!  real, dimension (:,:,:), allocatable :: vnew
  real, dimension (:,:), allocatable :: vnew!, vnew_ss
  ! (negrid,nspec) replicated
! <TT
  ! (naky,negrid,nspec) replicated

  ! MAB: new collision frequencies needed when energy diffusion included 
  ! vnew_ss can now be eliminated
  real, dimension (:,:), allocatable :: vnew_s, vnew_D, vnew_E

  ! TT: Do we want to keep ig dependence for vnewh?
  ! only for hyper-diffusive collisions
  real, dimension (:,:,:,:), allocatable :: vnewh
  ! (-ntgrid:ntgrid,ntheta0,naky,nspec) replicated

! TT> we can probably remove the first index -- Do they have to be complex?
  ! only for "new" momentum conservation (8.06)
!  complex, dimension(:,:,:), allocatable :: z0, z1
  real, dimension (:), allocatable :: z0
  real, dimension (:,:), allocatable :: z1
  ! TT: z0 may not need the first dimension
! <TT

  ! MAB>
  ! needed for momentum and energy conservation due to energy diffusion (3.08)
! TT> Can they be real?
!  complex, dimension(:,:,:), allocatable :: bs0, bw0, bz0
!  complex, dimension(:,:,:,:), allocatable :: v2w1
  complex, dimension(:), allocatable :: bs0
  complex, dimension(:,:), allocatable :: bw0, bz0
  complex, dimension(:,:,:), allocatable :: v2w1
! <TT
  ! <MAB

! TT> doesn't appear
  ! only for original parallel mom conservation (not used nowadays)
!  real, dimension (:,:,:), allocatable :: sq
  ! (-ntgrid:ntgrid,nlambda,2) replicated
! <TT

  ! MAB>
  ! only for energy diffusion
  real, dimension (:,:), allocatable :: ec1, ebetaa, eql
  complex, dimension (:,:), allocatable :: ged
  ! <MAB

  ! only for lorentz
  real :: cfac
  real, dimension (:,:), allocatable :: c1, betaa, ql, d1, h1
  complex, dimension (:,:), allocatable :: glz, glzc
  ! ( (2*nlambda+1), -*- lz_layout -*-)
  ! (-ntgrid:ntgrid,naky,2*nlambda+1,negrid,nspec)

  type (redist_type), save :: lorentz_map
  type (redist_type), save :: ediffuse_map  ! MAB

  logical :: hypermult
  logical :: initialized = .false.
  logical :: accelerated_x = .false.
  logical :: accelerated_v = .false.

contains

  subroutine init_collisions
    use species, only: init_species, nspec, spec
    use theta_grid, only: init_theta_grid, ntgrid
    use kgrids, only: init_kgrids, naky, nakx
    use le_grids, only: init_le_grids, nlambda, negrid 
    use run_parameters, only: init_run_parameters
    use agk_layouts, only: init_dist_fn_layouts, init_agk_layouts
    implicit none

    if (initialized) return
    initialized = .true.

    call init_agk_layouts
    call init_species

    hyper_colls = .false.
    if (any(spec%nu_h > epsilon(0.0))) hyper_colls = .true.

    call init_theta_grid
    call init_kgrids
    call init_le_grids (accelerated_x, accelerated_v)
    call init_run_parameters
    call init_dist_fn_layouts (ntgrid, naky, nakx, nlambda, negrid, nspec)

    call read_parameters
    call init_arrays
  end subroutine init_collisions

  subroutine read_parameters
    use file_utils, only: input_unit, error_unit, input_unit_exist
    use text_options
    use mp, only: proc0, broadcast
    implicit none
! MAB>
!    type (text_option), dimension (5), parameter :: modelopts = &
    type (text_option), dimension (6), parameter :: modelopts = &
!         (/ text_option('default', collision_model_lorentz), &
         (/ text_option('default', collision_model_full), &
            text_option('lorentz', collision_model_lorentz), &
            text_option('lorentz-test', collision_model_lorentz_test), &
            text_option('ediffuse', collision_model_ediffuse), &  ! MAB
            text_option('none', collision_model_none), &
            text_option('collisionless', collision_model_none) /)
! <MAB
    character(20) :: collision_model
    namelist /collisions_knobs/ collision_model, conserve_momentum, heating, adjust, const_v, cfac, hypermult, conserve_moments, test
    integer :: ierr, in_file
    logical :: exist

    if (proc0) then
       hypermult = .false.
       cfac = 1.   ! DEFAULT CHANGED TO INCLUDE CLASSICAL DIFFUSION: APRIL 18, 2006
       adjust = .true.
! TT>
       collision_model = 'default'  ! DEFAULT USES PITCH-ANGLE + ENERGY DIFFUSION
!       conserve_momentum = .true.  ! DEFAULT CHANGED TO REFLECT IMPROVED MOMENTUM CONSERVATION, 8/06
       conserve_momentum = .false.  ! DEFAULT CHANGED TO REFLECT ENERGY DIFFUSION OPERATOR, 4/08
! <TT
       conserve_moments = .true.   ! MAB
       const_v = .false.
       heating = .false.
       test = .false.
       in_file = input_unit_exist ("collisions_knobs", exist)
       if (exist) read (unit=input_unit("collisions_knobs"), nml=collisions_knobs)

       ierr = error_unit()
       call get_option_value &
            (collision_model, modelopts, collision_model_switch, &
            ierr, "collision_model in collisions_knobs")
    end if

! TT>
    if (conserve_momentum .and. &
         collision_model_switch /= collision_model_none) then
       ierr = error_unit()
       write (ierr,*) 'WARNING: conserve_momentum flag is obsolete!'
       write (ierr,*) 'WARNING: Now that energy diffusion operator is implemented (4/1/08),'
       write (ierr,*) 'WARNING: we should use conserve_moments option instead of conserve_momentum only!'
       conserve_moments = conserve_momentum
    end if
! <TT

    call broadcast (hypermult)
    call broadcast (cfac)
    call broadcast (conserve_momentum)
    call broadcast (conserve_moments)    ! MAB
    call broadcast (const_v)
    call broadcast (collision_model_switch)
    call broadcast (heating)
    call broadcast (adjust)
    call broadcast (test)
  end subroutine read_parameters

  subroutine init_arrays
    use species, only: nspec
    use le_grids, only: negrid
    use agk_layouts, only: g_lo
    use kgrids, only: naky, nakx
    use theta_grid, only: ntgrid
    use dist_fn_arrays, only: c_rate
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex
    implicit none
    real, dimension (negrid,nspec) :: hee
    logical :: first_time = .true.

    if (first_time) then
       call attempt_alloc("c_rate",tcomplex,5,2*ntgrid+1, nakx, naky, nspec, 3)
       allocate (c_rate(-ntgrid:ntgrid, nakx, naky, nspec, 3))
       call complete_alloc("c_rate")
       c_rate = 0.
       first_time = .false.
    end if

    if (collision_model_switch == collision_model_none) return

    call init_vnew (hee)
    if (all(abs(vnew(1,:)) <= 2.0*epsilon(0.0))) then
       collision_model_switch = collision_model_none
       return
    end if

    select case (collision_model_switch)
! MAB>
    case (collision_model_full)
       call init_lorentz
       call init_ediffuse
       if (conserve_moments) then
          call init_lorentz_conserve
          call init_diffuse_conserve
       end if
! <MAB
    case (collision_model_lorentz,collision_model_lorentz_test)
       call init_lorentz
!       if (conserve_momentum) call init_lz_mom_conserve
       if (conserve_moments) call init_lorentz_conserve    ! MAB
! MAB>
    case (collision_model_ediffuse)
       call init_ediffuse
       if (conserve_moments) call init_diffuse_conserve
! <MAB
    end select

  end subroutine init_arrays

!  subroutine init_lz_mom_conserve 
  subroutine init_lorentz_conserve   ! MAB
!
! Precompute two quantities needed for momentum conservation:
! z0, z1
!
    use agk_layouts, only: g_lo, ie_idx, is_idx, ik_idx, it_idx
    use species, only: nspec, spec
    use kgrids, only: naky, nakx
    use theta_grid, only: ntgrid
    use le_grids, only: e, integrate_moment
    use agk_time, only: dtime
    use dist_fn_arrays, only: aj0, aj1vp2, kperp2, vpa
    use agk_mem, only: attempt_alloc,complete_alloc,mem_dealloc,treal
    implicit none

    logical, save :: first = .true.
    complex, dimension (1,1,1) :: dum1 = 0., dum2 = 0.
! TT>
!    complex, dimension (:,:,:), allocatable :: gtmp
    real, dimension (:,:), allocatable :: gtmp
!    complex, dimension (:,:,:,:), allocatable :: v0z0, v0z1, v1z0, v1z1, Euinv
!    complex, dimension (:,:,:,:), allocatable :: v0z0, v0z1, v1z0, v1z1, duinv  ! MAB
    real, dimension (:,:,:), allocatable :: v0z0, v1z1, duinv
    complex, dimension (:,:,:), allocatable :: z_big
! <TT
    integer :: ie, ik, is, isgn, iglo, all, it

    if (first) then
! TT>
!       allocate (z0(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
!       allocate (z1(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
       call attempt_alloc("z0",treal,1,g_lo%ulim_alloc-g_lo%llim_proc+1)
       allocate (z0(g_lo%llim_proc:g_lo%ulim_alloc))
       call complete_alloc("z0")
       call attempt_alloc("z1",treal,2,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
       allocate (z1(2,g_lo%llim_proc:g_lo%ulim_alloc))
       call complete_alloc("z1")
! <TT
       first = .false.
    end if

! Enu should be replaced with du -- MAB
! First, get Enu and then 1/Enu == Enuinv

    call attempt_alloc("gtmp",treal,2,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
    allocate (gtmp(2,g_lo%llim_proc:g_lo%ulim_alloc))
    call complete_alloc("gtmp")
    call attempt_alloc("duinv",treal,3,nakx, naky, nspec)
    allocate (duinv(nakx, naky, nspec))
    call complete_alloc("duinv")

! Enu -> du, nu -> nu_D
!
! Enu == int (E nu f_0);  Enu = Enu(z, kx, ky, s)
! Enuinv = 1/Enu
    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       gtmp(:,iglo) = e(ie,is) * vnew_D(ie,is)
    end do

! No real z dependence anymore, since B is constant.  Could be simplified.
! TT: done

    all = 1
! MAB>
    call integrate_moment (gtmp, duinv, all)  ! not 1/du yet   

    where (abs(duinv) > epsilon(0.0))  ! necessary b/c some species may have nu=0
                                        ! duinv=0 iff vnew=0 so ok to keep duinv=0.
       duinv = 1./duinv  ! now it is 1/du
! <MAB
    end where

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get z0 & z1 (first form)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       it = it_idx(g_lo,iglo)
       ik = ik_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       z0(iglo) = - 1.5*vnew_D(ie,is)*aj1vp2(iglo) &
            * dtime * spec(is)%smz**2 * kperp2(it,ik) * duinv(it,ik,is)
       z1(:,iglo) = - 3.*vnew_D(ie,is)*vpa(:,iglo)*aj0(iglo) &
            * dtime * duinv(it,ik,is)
    end do

! MAB>
!    deallocate (Enuinv)  ! Done with this variable
    call mem_dealloc("duinv",treal,3,nakx, naky, nspec)
    deallocate (duinv)  ! Done with this variable
! <MAB

! TT> hack for the use of the general solver
!    call solfp_lorentz (z0, dum1, dum2)   ! z0 is redefined below
!    call solfp_lorentz (z1, dum1, dum2)   ! z1 is redefined below
    allocate (z_big(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
!    z_big = spread (spread(cmplx(z0,0.0), 1, 2), 1, ntgrid*2+1)
    z_big = 0.0
    z_big(0,:,:) = spread (cmplx(z0,0.0), 1, 2)
    z_big(1,:,:) = cmplx(z1,0.0)
    call solfp_lorentz (z_big, dum1, dum2)
    z0 = real(z_big(0,1,:))                ! z0 is redefined below
    z1 = real(z_big(1,:,:))                ! z1 is redefined below
    deallocate (z_big)
! <TT

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get v0z0

    call attempt_alloc("v0z0",treal,3,nakx, naky, nspec)
    allocate (v0z0(nakx, naky, nspec))
    call complete_alloc("v0z0")

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       gtmp(:,iglo) = 0.5*vnew_D(ie,is)*aj1vp2(iglo) * z0(iglo)
    end do

    call integrate_moment (gtmp, v0z0, all)    ! v0z0

! TT: v1z0 & v0z1 are zero due to linearity w.r.t vpar.
! TT we may omit some of the following steps.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get v1z1

    call attempt_alloc("v1z1",treal,3,nakx, naky, nspec)
    allocate (v1z1(nakx, naky, nspec))
    call complete_alloc("v1z1")

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       gtmp(:,iglo) = vnew_D(ie,is)*vpa(:,iglo)*aj0(iglo) * z1(:,iglo)
    end do

    call integrate_moment (gtmp, v1z1, all)    ! redefined below

    call mem_dealloc("gtmp",treal,2,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
    deallocate (gtmp)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now redefine z0 == z0/(1 + v0z0), z1 == z1/(1 + v1z1)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       z0(iglo) = z0(iglo) / (1. + v0z0(it,ik,is))
       z1(:,iglo) = z1(:,iglo) / (1. + v1z1(it,ik,is))
    end do

    call mem_dealloc("v0z0",treal,3,nakx, naky, nspec)
    call mem_dealloc("v1z1",treal,3,nakx, naky, nspec)
    deallocate (v0z0, v1z1)

! TT>
    if (test) then
       print*, 'writing out z0 & z1 in init_lorentz_conserve'
       print*, 'isign, iglo, z0, z1'
       do isgn = 1, 2
          do iglo = g_lo%llim_proc, g_lo%ulim_proc
             print*, isgn, iglo, z0(iglo), z1(isgn,iglo)
          end do
       end do
       stop
    end if
! <TT

!  end subroutine init_lz_mom_conserve
  end subroutine init_lorentz_conserve  ! MAB

! MAB>
  subroutine init_diffuse_conserve
!
! Precompute four quantities needed for momentum and energy conservation:
! bz0, bw0, bs0, v2w1
!
    use agk_layouts, only: g_lo, ie_idx, is_idx, ik_idx, it_idx
    use species, only: nspec, spec
    use kgrids, only: naky, nakx
    use theta_grid, only: ntgrid
    use le_grids, only: e, al, integrate_moment, negrid
    use agk_time, only: dtime
    use dist_fn_arrays, only: aj0, aj1vp2, kperp2, vpa

    logical, save :: first = .true.
! TT> Can they be real except bz_big?
!    complex, dimension (:,:,:), allocatable :: gtmp
!    complex, dimension (:,:,:,:), allocatable :: v0z0, duinv
    complex, dimension (:,:), allocatable :: gtmp
    complex, dimension (:,:,:), allocatable :: v0z0, duinv
    complex, dimension (:,:,:), allocatable :: bz_big
! <TT
    real, dimension (:,:), allocatable :: vns
    integer :: ie, il, ik, is, ig, isgn, iglo, all, it

    if (first) then
! TT>
!!$       allocate (bz0(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
!!$       allocate (bw0(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
!!$       allocate (bs0(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
!!$       allocate (v2w1(-ntgrid:ntgrid,nakx,naky,nspec))
       allocate (bz0(2,g_lo%llim_proc:g_lo%ulim_alloc))
       allocate (bw0(2,g_lo%llim_proc:g_lo%ulim_alloc))
       allocate (bs0(g_lo%llim_proc:g_lo%ulim_alloc))
       ! TT: bs0 may not need nsign dimension
       allocate (v2w1(nakx,naky,nspec))
! <TT
       first = .false.
    end if

! First, get du and then 1/du == duinv

! TT>
!    allocate (gtmp(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
!    allocate (duinv(-ntgrid:ntgrid, nakx, naky, nspec))       
    allocate (gtmp(2,g_lo%llim_proc:g_lo%ulim_alloc))
    allocate (duinv(nakx, naky, nspec))       
! <TT
    allocate (vns(negrid,nspec))

    vns = vnew_s - vnew_D

!
! du == int (E nu_s f_0);  du = du(z, kx, ky, s)
! duinv = 1/du
    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
!             gtmp(ig,isgn,iglo) = e(ie,is)*vns(ie,is)
       gtmp(:,iglo) = e(ie,is)*vns(ie,is)
!          end do
!       end do
! <TT
    end do

    all = 1
    call integrate_moment (gtmp, duinv, all)  ! not 1/du yet

    where (cabs(duinv) > epsilon(0.0))  ! necessary b/c some species may have vnewk=0
                                   ! duinv=0 iff vnew=0 so ok to keep duinv=0.
       duinv = 1./duinv  ! now it is 1/du
    end where

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get s0 (first form)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       it = it_idx(g_lo,iglo)
       ik = ik_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! V_perp == e(ie,is)*al(il)*aj1(ig,iglo)
! recall that aj1 == J1 (arg) / arg, arg = sqrt(kperp2*e*al*(T*m/B/q**2))
! u0 = -3 nu_s V_perp dt a f_0 / du
! where a = kperp2 * (T m / q**2) / b(theta)**2
!>MAB
! note that u0 here is defined slightly differently from u0 in write-up.
! in particular, the u0 from write-up has only factor of sqrt(a)...
! the additional factor of sqrt(a) appearing here is taken from v0
!             bs0(ig,isgn,iglo) = - 1.5*vns(ie,is)*aj1vp2(iglo) &
!                  * dtime * spec(is)%smz**2 * kperp2(it,ik) * duinv(ig,it,ik,is)
!<MAB
       bs0(iglo) = - 1.5*vns(ie,is)*aj1vp2(iglo) &
            * dtime * spec(is)%smz**2 * kperp2(it,ik) * duinv(it,ik,is)
!          end do
!       end do
! <TT
    end do

! TT> hack for the use of the general solver
!    call solfp_ediffuse (bs0)   ! s0 is redefined below
!    allocate (bz_big(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
!    bz_big = spread (spread(bs0, 1, 2), 1, ntgrid*2+1)
!    call solfp_ediffuse (bz_big)   ! s0 is redefined below
!    bs0 = bz_big(0,1,:)
! <TT

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get w0 (first form)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       it = it_idx(g_lo,iglo)
       ik = ik_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! v_parallel == vpa 
! V_parallel == v_parallel J0
! u1 = -3 nu V_parallel dt f_0 / du
!
!             bw0(ig,isgn,iglo) = - 3.*vns(ie,is)*vpa(isgn,iglo)*aj0(iglo) &
!                  * dtime * duinv(ig,it,ik,is)
       bw0(:,iglo) = - 3.*vns(ie,is)*vpa(:,iglo)*aj0(iglo) &
            * dtime * duinv(it,ik,is)
!          end do
!       end do
! <TT
    end do

! TT> hack for the use of the general solver
!    call solfp_ediffuse (bw0)    ! w0 is redefined below
!    bz_big = spread (bw0, 1, ntgrid*2+1)
!    call solfp_ediffuse (bz_big)    ! w0 is redefined below
!    bw0 = bz_big(0,:,:)
! <TT

! redefine du = dq (for energy-conserving term)
! du == int (E^2 nu_E f_0);  du = du(z, kx, ky, s)
! duinv = 1/du
    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
!             gtmp(ig,isgn,iglo)  = e(ie,is)**2*vnew_E(ie,is)
       gtmp(:,iglo)  = e(ie,is)**2 * vnew_E(ie,is)
!          end do
!       end do
! <TT
    end do

    all = 1
    call integrate_moment (gtmp, duinv, all)  ! not 1/du yet

    where (cabs(duinv) > epsilon(0.0))  ! necessary b/c some species may have vnewk=0
                                   ! duinv=0 iff vnew=0 so ok to keep duinv=0.
       duinv = 1./duinv  ! now it is 1/du
    end where

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get z0 (first form)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       it = it_idx(g_lo,iglo)
       ik = ik_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! u2 = -nu_E v^2 dt J0 f_0 / du
!             bz0(ig,isgn,iglo) = - vnew_E(ie,is)*e(ie,is)*aj0(iglo) &
!                  * dtime * duinv(ig,it,ik,is)
       bz0(:,iglo) = - vnew_E(ie,is)*e(ie,is)*aj0(iglo) &
            * dtime * duinv(it,ik,is)
!          end do
!       end do
! <TT
    end do

    deallocate (duinv)  ! Done with this variable

! TT> hack for the use of the general solver
!    call solfp_ediffuse (bz0)    ! z0 is redefined below
!    bz_big = spread (bz0, 1, ntgrid*2+1)
    allocate (bz_big(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc))
    bz_big = 0.0
    bz_big(0,:,:) = spread (bs0, 1, 2)
    bz_big(1,:,:) = bw0
    bz_big(-1,:,:) = bz0
    call solfp_ediffuse (bz_big)    ! z0 is redefined below
    bs0 = bz_big(0,1,:)
    bw0 = bz_big(1,:,:)
    bz0 = bz_big(-1,:,:)
    deallocate (bz_big)
! <TT

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get v2w1 (will redefine later...really getting v0s0 right now)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
! TT> unused
!       ik = ik_idx(g_lo,iglo)
! <TT
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! V_perp == e(ie,is)*al(il)*aj1(ig,iglo)
! v0 = nu V_perp
! no sqrt(kperp2 smz**2) because already absorbed into bs0 (u0 from write-up)
!             gtmp(ig,isgn,iglo) = 0.5*vns(ie,is)*aj1vp2(iglo) &
!                  * bs0(ig,isgn,iglo)
       gtmp(:,iglo) = 0.5*vns(ie,is)*aj1vp2(iglo) * bs0(iglo)
!          end do
!       end do
! <TT
    end do

    call integrate_moment (gtmp, v2w1, all)    ! v0s0

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine s0 = s0 / (1 + v0s0)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn=1,2
!          do ig=-ntgrid,ntgrid
!             bs0(ig,isgn,iglo) = bs0(ig,isgn,iglo) / (1.0 + v2w1(ig,it,ik,is))
       bs0(iglo) = bs0(iglo) / (1.0 + v2w1(it,ik,is))
!          end do
!       end do
! <TT
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine v2w1 to actually be v0w0 (instead of v0s0)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
! TT> unused
!       ik = ik_idx(g_lo,iglo)
! <TT
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! V_perp == e(ie,is)*al(il)*aj1(ig,iglo)
! v0 = nu V_perp
! no sqrt(kperp2 smz**2) because already absorbed into s0 (u0 from write-up)
!             gtmp(ig,isgn,iglo) = 0.5*vns(ie,is)*aj1vp2(iglo) &
!                  * bw0(ig,isgn,iglo)
       gtmp(:,iglo) = 0.5*vns(ie,is)*aj1vp2(iglo) * bw0(:,iglo)
!          end do
!       end do
! <TT
    end do

    call integrate_moment (gtmp, v2w1, all)    ! v0w0

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine w0 = w0 - v0w0 * s0 (this is w1 from MAB notes)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn=1,2
!          do ig=-ntgrid,ntgrid
!             bw0(ig,isgn,iglo) = bw0(ig,isgn,iglo) - bs0(ig,isgn,iglo)*v2w1(ig,it,ik,is)
       bw0(:,iglo) = bw0(:,iglo) - bs0(iglo)*v2w1(it,ik,is)
!          end do
!       end do
! <TT
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get v0z0

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
! TT> unused
!       ik = ik_idx(g_lo,iglo)
! <TT
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! V_perp == e(ie,is)*al(il)*aj1(ig,iglo)
! v0 = nu V_perp
! no sqrt(kperp2 smz**2) because already absorbed into s0 (u0 from write-up)
!             gtmp(ig,isgn,iglo) = 0.5*vns(ie,is)*aj1vp2(iglo) &
!                  * bz0(ig,isgn,iglo)
       gtmp(:,iglo) = 0.5*vns(ie,is)*aj1vp2(iglo) * bz0(:,iglo)
!          end do
!       end do
! <TT
    end do

! TT>
!    allocate (v0z0(-ntgrid:ntgrid,nakx,naky,nspec))
    allocate (v0z0(nakx,naky,nspec))
! <TT
    call integrate_moment (gtmp, v0z0, all)    ! v0z0

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine z0 = z0 - v0z0 * s0 (this is z1 from MAB notes)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn=1,2
!          do ig=-ntgrid,ntgrid
!             bz0(ig,isgn,iglo) = bz0(ig,isgn,iglo) - bs0(ig,isgn,iglo)*v0z0(ig,it,ik,is)
       bz0(:,iglo) = bz0(:,iglo) - bs0(iglo)*v0z0(it,ik,is)
!          end do
!       end do
! <TT
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine v2w1 = v1 . w1

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! v_parallel == vpa
! V_parallel == v_parallel J0
! v1 = nu V_parallel f_0
!
!             gtmp(ig,isgn,iglo) = vns(ie,is)*vpa(isgn,iglo)*aj0(iglo) &
!                  * bw0(ig,isgn,iglo)
       gtmp(:,iglo) = vns(ie,is)*vpa(:,iglo)*aj0(iglo) * bw0(:,iglo)
!          end do
!       end do
! <TT
    end do

    call integrate_moment (gtmp, v2w1, all)    ! v1w1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine w0 = w1 / (1 + v1 . w1)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn=1,2
!          do ig=-ntgrid,ntgrid
!             bw0(ig,isgn,iglo) = bw0(ig,isgn,iglo) / (1.0 + v2w1(ig,it,ik,is))
       bw0(:,iglo) = bw0(:,iglo) / (1.0 + v2w1(it,ik,is))
!          end do
!       end do
! <TT
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine v0z0 = v1 . z1

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT>
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! v_parallel == vpa
! V_parallel == v_parallel J0
! v1 = nu V_parallel f_0
!
!             gtmp(ig,isgn,iglo) = vns(ie,is)*vpa(isgn,iglo)*aj0(iglo) &
!                  * bz0(ig,isgn,iglo)
       gtmp(:,iglo) = vns(ie,is)*vpa(:,iglo)*aj0(iglo) * bz0(:,iglo)
!          end do
!       end do
! <TT
    end do

    deallocate (vns)

    call integrate_moment (gtmp, v0z0, all)    ! v1z1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine z0 = z1 - v1z1 * w1 (this is z2 from MAB notes)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn=1,2
!          do ig=-ntgrid,ntgrid
!             bz0(ig,isgn,iglo) = bz0(ig,isgn,iglo) - bw0(ig,isgn,iglo)*v0z0(ig,it,ik,is)
       bz0(:,iglo) = bz0(:,iglo) - bw0(:,iglo)*v0z0(it,ik,is)
!          end do
!       end do
! <TT
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine v2w1 = v2 . w1

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! v2 = nu_E*e*J0 
!             gtmp(ig,isgn,iglo) = vnew_E(ie,is)*e(ie,is)*aj0(iglo) &
!                  * bw0(ig,isgn,iglo)
       gtmp(:,iglo) = vnew_E(ie,is)*e(ie,is)*aj0(iglo) * bw0(:,iglo)
!          end do
!       end do
! <TT
    end do

    call integrate_moment (gtmp, v2w1, all)   ! v2w1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Redefine v0z0 = v2 . z2

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! v2 = nu_E*e*J0 
!             gtmp(ig,isgn,iglo) = vnew_E(ie,is)*e(ie,is)*aj0(iglo) &
!                  * bz0(ig,isgn,iglo)
       gtmp(:,iglo) = vnew_E(ie,is)*e(ie,is)*aj0(iglo) * bz0(:,iglo)
!          end do
!       end do
! <TT
    end do

    call integrate_moment (gtmp, v0z0, all)   ! v2z2

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now redefine z0 = z2 / (1 + v2z2)

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT> reduced dimension
!       do isgn=1,2
!          do ig=-ntgrid,ntgrid
!             bz0(ig,isgn,iglo) = bz0(ig,isgn,iglo) / (1.0 + v0z0(ig,it,ik,is))
       bz0(:,iglo) = bz0(:,iglo) / (1.0 + v0z0(it,ik,is))
!          end do
!       end do
! <TT
    end do

    deallocate (v0z0)

  end subroutine init_diffuse_conserve

  subroutine init_vnew (hee)
    use species, only: nspec, spec, electron_species
    use le_grids, only: negrid, e
    use kgrids, only: naky, nakx
    use theta_grid, only: ntgrid
    use run_parameters, only: zeff
    use dist_fn_arrays, only: kperp2
    use constants
    use agk_mem, only: attempt_alloc,complete_alloc,treal,mem_dealloc
    implicit none
    real, dimension (:,:), intent (out) :: hee
!    real,dimension (negrid,nspec)::heevth
    real,dimension (negrid,nspec) :: heevth, hsg, hsgvth  ! MAB
    integer :: ik, ie, is, it, ig
    real :: v, k_nexp_h_max 

    do is = 1, nspec
       do ie = 1, negrid
          v = sqrt(e(ie,is))
          hee(ie,is) = 1.0/sqrt(pi)/v*exp(-e(ie,is)) &
               + (1.0 - 0.5/e(ie,is)) &
               *(1.0 - 1.0/(1.0          + v &
               *(0.0705230784 + v &
               *(0.0422820123 + v &
               *(0.0092705272 + v &
               *(0.0001520143 + v &
               *(0.0002765672 + v &
               *(0.0000430638)))))))**16)

! MAB>
! hsg is the G of Hirshman and Sigmar
! added to allow for momentum conservation with energy diffusion
          hsg(ie,is) = -1.0/sqrt(pi)/v*exp(-e(ie,is)) &
               + 0.5/e(ie,is) &
               *(1.0 - 1.0/(1.0 + v &  ! this line on is erf(v)
               *(0.0705230784 + v &
               *(0.0422820123 + v &
               *(0.0092705272 + v &
               *(0.0001520143 + v &
               *(0.0002765672 + v &
               *(0.0000430638)))))))**16)
! <MAB
       end do
    end do

!heevth is hee but using only thermal velocity (energy independent)

    do is = 1, nspec
       do ie = 1, negrid
          v = 1           
          heevth(ie,is) = 1.0/sqrt(pi)/v*exp(-v**2) &
               + (1.0 - 0.5/v**2) &
               *(1.0 - 1.0/(1.0          + v &
               *(0.0705230784 + v &
               *(0.0422820123 + v &
               *(0.0092705272 + v &
               *(0.0001520143 + v &
               *(0.0002765672 + v &
               *(0.0000430638)))))))**16)

! MAB>
! hsg is the G of Helander and Sigmar
! added to allow for momentum conservation with energy diffusion
          hsgvth(ie,is) = -1.0/sqrt(pi)/v*exp(-v**2) &
               + 0.5/v**2 &
               *(1.0 - 1.0/(1.0 + v &
               *(0.0705230784 + v &
               *(0.0422820123 + v &
               *(0.0092705272 + v &
               *(0.0001520143 + v &
               *(0.0002765672 + v &
               *(0.0000430638)))))))**16)
! <MAB
       end do
    end do                                                                  

!    do is = 1, nspec
!       if (spec(is) % nustar < 0.)
!
!    end do

! TT>
!    if(.not.allocated(vnew)) allocate (vnew(naky,negrid,nspec))
    if (.not.allocated(vnew)) then
       allocate (vnew(negrid,nspec))
! MAB>
!       allocate (vnew_ss(negrid,nspec))
       allocate (vnew_s(negrid,nspec))
       allocate (vnew_D(negrid,nspec))
       allocate (vnew_E(negrid,nspec))
! <MAB
    end if
! <TT
    if(.not.allocated(vnewh)) then
       call attempt_alloc("vnewh",treal,4,2*ntgrid+1,nakx, naky, nspec)
       allocate (vnewh(-ntgrid:ntgrid,nakx,naky,nspec))
       call complete_alloc("vnewh")
    endif

    do is = 1, nspec
       if (spec(is)%type == electron_species) then
          do ie = 1, negrid
             if (const_v) then
! TT>
!                   vnew(ik,ie,is) = spec(is)%nu * (zeff + heevth(ie,is)) * 0.5
                vnew(ie,is) = spec(is)%nu * (zeff + heevth(ie,is)) * 0.5
! MAB>
!                   vnew_ss(ie,is) = spec(is)%nu * heevth(ie,is) * 0.5
                vnew_s(ie,is) = spec(is)%nu * hsgvth(ie,is) * 4.0
                vnew_D(ie,is) = spec(is)%nu * heevth(ie,is)
                vnew_E(ie,is) = vnew_s(ie,is) * 1.5 - 2.0 * vnew_D(ie,is)
! <MAB
! <TT
             else
! TT>
!                   vnew(ik,ie,is) = spec(is)%nu / e(ie,is)**1.5 &
                vnew(ie,is) = spec(is)%nu / e(ie,is)**1.5 &
                     *(zeff + hee(ie,is))*0.5        ! w/ ee
!                        * zeff * 0.5                    ! no ee
! MAB>
!                   vnew_ss(ie,is) = spec(is)%nu / e(ie,is)**1.5 &
!                        * hee(ie,is) * 0.5
                vnew_s(ie,is) = spec(is)%nu / sqrt(e(ie,is)) * hsg(ie,is) * 4.0
                vnew_D(ie,is) = spec(is)%nu / e(ie,is)**1.5 * hee(ie,is)
                vnew_E(ie,is) = vnew_s(ie,is) * (2.0 - 0.5 / e(ie,is)) &
                     - 2.0 * vnew_D(ie,is)
! <MAB
! <TT
             end if
          end do
       else
          do ie = 1, negrid
             if (const_v) then
! TT>
!                   vnew(ik,ie,is) = spec(is)%nu * heevth(ie,is) * 0.5
                vnew(ie,is) = spec(is)%nu * heevth(ie,is) * 0.5
! <TT
             else
! TT>
!                   vnew(ik,ie,is) = spec(is)%nu / e(ie,is)**1.5 &
                vnew(ie,is) = spec(is)%nu / e(ie,is)**1.5 &
! <TT
                     * hee(ie,is) * 0.5
             end if
! TT>
             ! factor of 2 may be needed?
! MAB>
!                vnew_ss(ie,is) = vnew(ie,is)
             vnew_s(ie,is) = spec(is)%nu / sqrt(e(ie,is)) * hsg(ie,is) * 4.0
             vnew_D(ie,is) = 2.0 * vnew(ie,is)
             vnew_E(ie,is) = vnew_s(ie,is) * (2.0 - 0.5 / e(ie,is)) &
                  - 2.0 * vnew_D(ie,is)
! <MAB
! <TT
          end do
       end if

       ! add hyper-terms inside collision operator
!BD: Warning!
!BD: Also: there is no "grid_norm" option here and the exponent is fixed to 4 for now
!GGH: 07SEP07- Fixed so that exponent is given by spec(is)%nexp_h (default=4.0)
       if (hyper_colls) then
          k_nexp_h_max = (maxval(kperp2))**(spec(is)%nexp_h/2.)
          do ik = 1, naky
             do it = 1, nakx
                do ig=-ntgrid,ntgrid
                   vnewh(ig,it,ik,is) = spec(is)%nu_h * kperp2(it,ik)**(spec(is)%nexp_h/2.)/k_nexp_h_max
                end do
             end do
          end do
       else
          vnewh = 0.
       end if
    end do

  end subroutine init_vnew
!================================================================================
! Update vnewh if adaptive hypercollisions has changed nu_h
! GGH: 2007 JUN 5
!================================================================================
  subroutine update_vnewh
    use species, only: nspec, spec, specie
    use kgrids, only: naky, nakx
    use theta_grid, only: ntgrid
    use le_grids, only: negrid
    use dist_fn_arrays, only: kperp2
    implicit none
    integer :: ik, is, it, ig
    real :: k_nexp_h_max            !kperp_max normalization of hypercollisionality
    real, dimension (negrid,nspec) :: hee

    !Recalculate vnewh based on updated value of nu_h
    do is = 1, nspec
       if (hyper_colls) then
          k_nexp_h_max = (maxval(kperp2))**(spec(is)%nexp_h/2.) 
          do ik = 1, naky
             do it = 1, nakx
                do ig=-ntgrid,ntgrid
                   vnewh(ig,it,ik,is) = spec(is)%nu_h * kperp2(it,ik)**(spec(is)%nexp_h/2.)/k_nexp_h_max
                end do
             end do
          end do
       else
          vnewh = 0.
       end if
    enddo
    
    !Recalculate arrays for collision operator
    select case (collision_model_switch)
    case (collision_model_lorentz,collision_model_lorentz_test)
       call init_lorentz
!       if (conserve_momentum) call init_lz_mom_conserve
       if (conserve_moments) call init_lorentz_conserve   ! MAB
    end select

  end subroutine update_vnewh
!================================================================================

  subroutine init_ediffuse

    use constants, only: pi
    use mp, only: proc0
    use species, only: nspec, spec
    use theta_grid, only: ntgrid
    use kgrids, only: naky, nakx
    use le_grids, only: nlambda, negrid, ng2, e, ecut, integrate_moment, al
    use egrid, only: zeroes, energy
    use agk_time, only: dtime
    use agk_layouts, only: init_ediffuse_layouts
    use agk_layouts, only: e_lo, g_lo, ik_idx, is_idx
    use agk_layouts, only: it_idx, il_idx
    use file_utils, only: open_output_file, close_output_file
    use dist_fn_arrays, only: kperp2

    implicit none

    integer :: ie, is, ik, ielo, il, it, ierr
    real, dimension (:), allocatable :: aa, bb, cc, xe
    real :: vn, xe0, xe1, xe2, xer, xel, er, el, fac
    real :: dela, delb, delc, delfac
    real :: capgl, capgr, slb1, ee

    call init_ediffuse_layouts &
         (ntgrid, naky, nakx, nlambda, nspec)
    call init_ediffuse_redistribute

    if (.not.allocated(ged)) then
       allocate (ged(negrid+1,e_lo%llim_proc:e_lo%ulim_alloc))
       ged = 0.0
    end if

    allocate (aa(negrid), bb(negrid), cc(negrid))
    allocate (xe(negrid))

! want to use x variables instead of e because we want conservative form
! for the x-integration
! TT> extended zeroes to have size negrid
!    xe(1:negrid-1) = zeroes
!    xe(negrid) = x0
    xe = zeroes
! <TT

    if (.not.allocated(ec1)) then
       allocate (ec1   (negrid,e_lo%llim_proc:e_lo%ulim_alloc))
       allocate (ebetaa(negrid,e_lo%llim_proc:e_lo%ulim_alloc))
       allocate (eql   (negrid,e_lo%llim_proc:e_lo%ulim_alloc))
    endif

    ec1 = 0.0 ; ebetaa = 0.0 ; eql = 0.0

    do ielo = e_lo%llim_proc, e_lo%ulim_proc
       is = is_idx(e_lo, ielo)
       ik = ik_idx(e_lo, ielo)
       il = il_idx(e_lo, ielo)
       it = it_idx(e_lo, ielo)

       vn = spec(is)%nu

       slb1 = sqrt(abs(1.0 - al(il)))     ! xi_j

       do ie = 2, negrid-1

          xe0 = xe(ie-1)
          xe1 = xe(ie)
          xe2 = xe(ie+1)

          capgl = 0.5*(xe0*sqrt(e(ie-1,is))*exp(-2.0*e(ie-1,is)) &
               +xe1*sqrt(e(ie,is))*exp(-2.0*e(ie,is)))
          capgr = 0.5*(xe2*sqrt(e(ie+1,is))*exp(-2.0*e(ie+1,is)) &
               +xe1*sqrt(e(ie,is))*exp(-2.0*e(ie,is)))

          dela = xe1 - xe0
          delb = xe2 - xe0
          delc = xe2 - xe1
          delfac = delc/dela - dela/delc
       
          fac = -16.0*vn*dtime/pi

          ee = 0.25/e(ie,is)**1.5*(1-slb1**2)*xe1 &
               / spec(is)%zstm**2 &
               * kperp2(it,ik)*cfac

          aa(ie) = fac*delc/(delb*dela)*(capgl/dela &
               - xe1*sqrt(e(ie,is))*exp(-2.0*e(ie,is))*delfac/delb)
          cc(ie) = fac*dela/(delb*delc)*(capgr/delc &
               + xe1*sqrt(e(ie,is))*exp(-2.0*e(ie,is))*delfac/delb)
          bb(ie) = 1.0 - (aa(ie) + cc(ie)) + ee*vn*dtime

       end do

       ! boundary at xe = 0
       xe1 = xe(1)
       xe2 = xe(2)

       xer = (xe2 + xe1)*0.5

       er = energy(xer,ecut,ierr)

       fac = -8.0*vn*dtime/pi

       ee = 0.25/e(1,is)**1.5*(1-slb1**2)*xe1 &
            / spec(is)%zstm**2 &
            * kperp2(it,ik)*cfac

       aa(1) = 0.0
       cc(1) = fac*sqrt(er)*exp(-2.0*er)/(xe2 - xe1)
       bb(1) = 1.0 - cc(1) + ee*vn*dtime

       ! boundary at xe = 1
       xe0 = xe(negrid-1)
       xe1 = xe(negrid)
       xe2 = 1.0

       xel = (xe1 + xe0)*0.5

       el = energy(xel,ecut,ierr)

       fac = -8.0*vn*dtime/pi/(xe2 - xel)

       ee = 0.25/e(negrid,is)**1.5*(1-slb1**2)*xe1 &
            / spec(is)%zstm**2 &
            * kperp2(it,ik)*cfac

       aa(negrid) = fac*sqrt(el)*xel*exp(-2.0*el)/(xe1-xe0)
       cc(negrid) = 0.0
       bb(negrid) = 1.0 - aa(negrid) + ee*vn*dtime

! fill in the arrays for the tridiagonal
       ec1(:,ielo) = cc
       ebetaa(1,ielo) = 1.0/bb(1)
       do ie = 1, negrid-1
          eql(ie+1,ielo) = aa(ie+1)*ebetaa(ie,ielo)
          ebetaa(ie+1,ielo) = 1.0/(bb(ie+1)-eql(ie+1,ielo)*ec1(ie,ielo))
       end do

    end do

    deallocate(aa, bb, cc, xe)

  end subroutine init_ediffuse

! TT>
  subroutine init_ediffuse_redistribute
    use mp, only: nproc
    use species, only: nspec
    use theta_grid, only: ntgrid
    use kgrids, only: naky, nakx
    use le_grids, only: nlambda, negrid, ng2
    use agk_layouts, only: init_ediffuse_layouts
    use agk_layouts, only: g_lo, e_lo, ie_idx, ik_idx, it_idx, il_idx, is_idx
    use agk_layouts, only: idx_local, proc_id, idx
    use redistribute, only: index_list_type, init_redist, delete_list
    implicit none
    type (index_list_type), dimension(0:nproc-1) :: to_list, from_list
    integer, dimension(0:nproc-1) :: nn_to, nn_from
    integer, dimension(3) :: from_low, from_high
    integer, dimension(2) :: to_high
    integer :: to_low
    integer :: ig, isign, iglo, ik, it, il, ie, is, ielo
    integer :: n, ip
    logical :: done = .false.

    if (done) return

    call init_ediffuse_layouts (ntgrid, naky, nakx, nlambda, nspec)

    ! count number of elements to be redistributed to/from each processor
    nn_to = 0
    nn_from = 0
    do iglo = g_lo%llim_world, g_lo%ulim_world
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       il = il_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       do isign = 1, 2
          do ig = -ntgrid, ntgrid
             ielo = idx(e_lo,ig,isign,ik,it,il,is)
             if (idx_local(g_lo,iglo)) &
                  nn_from(proc_id(e_lo,ielo)) = nn_from(proc_id(e_lo,ielo)) + 1
             if (idx_local(e_lo,ielo)) &
                  nn_to(proc_id(g_lo,iglo)) = nn_to(proc_id(g_lo,iglo)) + 1
          end do
       end do
    end do

    do ip = 0, nproc-1
       if (nn_from(ip) > 0) then
          allocate (from_list(ip)%first(nn_from(ip)))
          allocate (from_list(ip)%second(nn_from(ip)))
          allocate (from_list(ip)%third(nn_from(ip)))
       end if
       if (nn_to(ip) > 0) then
          allocate (to_list(ip)%first(nn_to(ip)))
          allocate (to_list(ip)%second(nn_to(ip)))
       end if
    end do

    ! get local indices of elements distributed to/from other processors
    nn_to = 0
    nn_from = 0
    do iglo = g_lo%llim_world, g_lo%ulim_world
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       il = il_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       do isign = 1, 2
          do ig = -ntgrid, ntgrid
             ielo = idx(e_lo,ig,isign,ik,it,il,is)
             if (idx_local(g_lo,iglo)) then
                ip = proc_id(e_lo,ielo)
                n = nn_from(ip) + 1
                nn_from(ip) = n
                from_list(ip)%first(n) = ig
                from_list(ip)%second(n) = isign
                from_list(ip)%third(n) = iglo
             end if
             if (idx_local(e_lo,ielo)) then
                ip = proc_id(g_lo,iglo)
                n = nn_to(ip) + 1
                nn_to(ip) = n
                to_list(ip)%first(n) = ie
                to_list(ip)%second(n) = ielo
             end if
          end do
       end do
    end do

    from_low (1) = -ntgrid
    from_low (2) = 1
    from_low (3) = g_lo%llim_proc

    to_low = e_lo%llim_proc

    to_high(1) = negrid+1
    to_high(2) = e_lo%ulim_alloc

    from_high(1) = ntgrid
    from_high(2) = 2
    from_high(3) = g_lo%ulim_alloc

    call init_redist (ediffuse_map, 'c', to_low, to_high, to_list, &
         from_low, from_high, from_list)

    call delete_list (to_list)
    call delete_list (from_list)

    done = .true.

  end subroutine init_ediffuse_redistribute
! <TT

  subroutine init_lorentz
    use species, only: nspec, spec
    use theta_grid, only: ntgrid
    use kgrids, only: naky, nakx
    use le_grids, only: nlambda, negrid, al, e
    use agk_time, only: dtime
    use dist_fn_arrays, only: kperp2
    use agk_layouts, only: init_lorentz_layouts
    use agk_layouts, only: lz_lo
    use agk_layouts, only: ig_idx, ik_idx, ie_idx, is_idx, it_idx
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex
    implicit none
    integer :: ig, il, ilz, it, ik, ie, is
    real, dimension (nlambda+1) :: aa, bb, cc, dd, hh
    real, dimension (2*nlambda+1) :: a1, b1
    real :: xi0, xi1, xi2, xil, xir, vn, ee, vnh, vnc

    call init_lorentz_layouts (ntgrid, naky, nakx, nlambda, negrid, nspec)
    call init_lorentz_redistribute

    if (.not.allocated(glz)) then
       call attempt_alloc("glz",tcomplex,2,2*nlambda+1,lz_lo%ulim_alloc-lz_lo%llim_proc+1)
       allocate (glz(2*nlambda+1,lz_lo%llim_proc:lz_lo%ulim_alloc))
       call complete_alloc("glz")
       glz = 0.0
       if (heating) then
          call attempt_alloc("glzc",tcomplex,2,2*nlambda+1,lz_lo%ulim_alloc-lz_lo%llim_proc+1)
          allocate (glzc(2*nlambda+1,lz_lo%llim_proc:lz_lo%ulim_alloc))
          call complete_alloc("glzc")
          glzc = 0.0
       end if
    end if

    if (.not.allocated(c1)) then
       call attempt_alloc("c1",tcomplex,2,2*nlambda+1,lz_lo%ulim_alloc-lz_lo%llim_proc+1)
       allocate (c1   (2*nlambda+1,lz_lo%llim_proc:lz_lo%ulim_alloc))
       call complete_alloc("c1")
       call attempt_alloc("betaa",tcomplex,2,2*nlambda+1,lz_lo%ulim_alloc-lz_lo%llim_proc+1)
       allocate (betaa(2*nlambda+1,lz_lo%llim_proc:lz_lo%ulim_alloc))
       call complete_alloc("betaa")
       call attempt_alloc("ql",tcomplex,2,2*nlambda+1,lz_lo%ulim_alloc-lz_lo%llim_proc+1)
       allocate (ql   (2*nlambda+1,lz_lo%llim_proc:lz_lo%ulim_alloc))
       call complete_alloc("ql")
       call attempt_alloc("d1",tcomplex,2,2*nlambda+1,lz_lo%ulim_alloc-lz_lo%llim_proc+1)
       allocate (d1   (2*nlambda+1,lz_lo%llim_proc:lz_lo%ulim_alloc))
       call complete_alloc("d1")
       call attempt_alloc("h1",tcomplex,2,2*nlambda+1,lz_lo%ulim_alloc-lz_lo%llim_proc+1)
       allocate (h1   (2*nlambda+1,lz_lo%llim_proc:lz_lo%ulim_alloc))
       call complete_alloc("h1")
    endif

    c1 = 0.0 ; betaa = 0.0 ; ql = 0.0 ; d1 = 0.0 ; h1 = 0.0

    do ilz = lz_lo%llim_proc, lz_lo%ulim_proc
       is = is_idx(lz_lo,ilz)
       ik = ik_idx(lz_lo,ilz)
       it = it_idx(lz_lo,ilz)
       ie = ie_idx(lz_lo,ilz)
       ig = ig_idx(lz_lo,ilz)
       if (collision_model_switch == collision_model_lorentz_test) then
          vn = abs(spec(is)%nu)
          vnc = 0.
          vnh = 0.
       else
          if (hypermult) then
! TT>
!             vn = vnew(ik,ie,is)*(1.+vnewh(ig,it,ik,is))
!             vnc = vnew(ik,ie,is)
             vn = vnew(ie,is)*(1.+vnewh(ig,it,ik,is))
             vnc = vnew(ie,is)
! <TT
             vnh = vnewh(ig,it,ik,is)*vnc
          else
! TT>
!             vn = vnew(ik,ie,is)+vnewh(ig,it,ik,is)
!             vnc = vnew(ik,ie,is)
             vn = vnew(ie,is)+vnewh(ig,it,ik,is)
             vnc = vnew(ie,is)
! <TT
             vnh = vnewh(ig,it,ik,is)
          end if
       end if

       do il = 2, nlambda-1
          ! xi = v_par/v
          xi0 = sqrt(abs(1.0 - al(il-1)))   ! xi_{j-1}
          xi1 = sqrt(abs(1.0 - al(il)))     ! xi_j
          xi2 = sqrt(abs(1.0 - al(il+1)))   ! xi_{j+1}

          xil = (xi1 + xi0)/2.0  ! xi(j-1/2)
          xir = (xi1 + xi2)/2.0  ! xi(j+1/2)

          ee = 0.5*e(ie,is)*(1+xi1**2) / spec(is)%zstm**2 * kperp2(it,ik)*cfac

          ! coefficients for tridiagonal matrix:
          cc(il) = -vn*dtime*(1.0 - xir*xir)/(xir - xil)/(xi2 - xi1)
          aa(il) = -vn*dtime*(1.0 - xil*xil)/(xir - xil)/(xi1 - xi0)
          bb(il) = 1.0 - (aa(il) + cc(il)) + ee*vn*dtime

          ! coefficients for entropy heating calculation
          dd(il) =vnc*((1.0-xir*xir)/(xir-xil)/(xi2-xi1) + ee)
          hh(il) =vnh*((1.0-xir*xir)/(xir-xil)/(xi2-xi1) + ee)
       end do

! boundary at xi = 1
       xi0 = 1.0
       xi1 = sqrt(abs(1.0-al(1)))
       xi2 = sqrt(abs(1.0-al(2)))

       xil = (xi1 + xi0)/2.0
       xir = (xi1 + xi2)/2.0

       ee = 0.5*e(ie,is)*(1+xi1**2) / spec(is)%zstm**2 * kperp2(it,ik)*cfac

       cc(1) = -vn*dtime*(-1.0 - xir)/(xi2-xi1)
       aa(1) = 0.0
       bb(1) = 1.0 - (aa(1) + cc(1)) + ee*vn*dtime

       dd(1) =vnc*((1.0-xir*xir)/(xir-xil)/(xi2-xi1) + ee)
       hh(1) =vnh*((1.0-xir*xir)/(xir-xil)/(xi2-xi1) + ee)

! boundary at xi = 0
       il = nlambda
       xi0 = sqrt(abs(1.0 - al(il-1)))
       xi1 = sqrt(abs(1.0 - al(il)))
       xi2 = -xi1

       xil = (xi1 + xi0)/2.0
       xir = (xi1 + xi2)/2.0

       ee = 0.5*e(ie,is)*(1+xi1**2) / spec(is)%zstm**2 * kperp2(it,ik)*cfac

       cc(il) = -vn*dtime*(1.0 - xir*xir)/(xir - xil)/(xi2 - xi1)
       aa(il) = -vn*dtime*(1.0 - xil*xil)/(xir - xil)/(xi1 - xi0)
       bb(il) = 1.0 - (aa(il) + cc(il)) + ee*vn*dtime

       dd(il) =vnc*((1.0-xir*xir)/(xir-xil)/(xi2-xi1) + ee)
       hh(il) =vnh*((1.0-xir*xir)/(xir-xil)/(xi2-xi1) + ee)

! start to fill in the arrays for the tridiagonal
       a1(:nlambda) = aa(:nlambda)
       b1(:nlambda) = bb(:nlambda)
       c1(:nlambda,ilz) = cc(:nlambda)

       d1(:nlambda,ilz) = dd(:nlambda)
       h1(:nlambda,ilz) = hh(:nlambda)

! assuming symmetry in xi, fill in the rest of the arrays.
       a1(nlambda+1:2*nlambda)     = cc(nlambda:1:-1)
       b1(nlambda+1:2*nlambda)     = bb(nlambda:1:-1)
       c1(nlambda+1:2*nlambda,ilz) = aa(nlambda:1:-1)

       d1(nlambda+1:2*nlambda,ilz) = dd(nlambda:1:-1)
       h1(nlambda+1:2*nlambda,ilz) = hh(nlambda:1:-1)

       betaa(1,ilz) = 1.0/b1(1)
       do il = 1, 2*nlambda-1
          ql(il+1,ilz) = a1(il+1)*betaa(il,ilz)
          betaa(il+1,ilz) = 1.0/(b1(il+1)-ql(il+1,ilz)*c1(il,ilz))
       end do

       ql(1,ilz) = 0.0

       ql(2*nlambda+1,ilz) = 0.0
       c1(2*nlambda+1,ilz) = 0.0
       betaa(2*nlambda+1,ilz) = 0.0

       d1(2*nlambda+1,ilz) = 0.0
       h1(2*nlambda+1,ilz) = 0.0
    end do
  end subroutine init_lorentz

  subroutine init_lorentz_redistribute
    use mp, only: nproc
    use species, only: nspec
    use theta_grid, only: ntgrid
    use kgrids, only: naky, nakx
    use le_grids, only: nlambda, negrid
    use agk_layouts, only: init_lorentz_layouts
    use agk_layouts, only: g_lo, lz_lo
    use agk_layouts, only: idx_local, proc_id
! TT>
!    use agk_layouts, only: gidx2lzidx 
    use agk_layouts, only: ik_idx, it_idx, ie_idx, is_idx, il_idx, idx
! <TT
    use redistribute, only: index_list_type, init_redist, delete_list
    implicit none
    type (index_list_type), dimension(0:nproc-1) :: to_list, from_list
    integer, dimension(0:nproc-1) :: nn_to, nn_from
    integer, dimension(3) :: from_low, from_high
    integer, dimension(2) :: to_high
    integer :: to_low
    integer :: ig, isign, iglo, il, ilz
! TT>
    integer :: ik, it, ie, is
! <TT
    integer :: n, ip
    logical :: done = .false.

    if (done) return

    call init_lorentz_layouts (ntgrid, naky, nakx, nlambda, negrid, nspec)

    ! count number of elements to be redistributed to/from each processor
    nn_to = 0
    nn_from = 0
    do iglo = g_lo%llim_world, g_lo%ulim_world
! TT> expansion of gidx2lzidx
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! <TT
       do isign = 1, 2
          do ig = -ntgrid, ntgrid
! TT>
!             call gidx2lzidx (ig, isign, g_lo, iglo, lz_lo, ntgrid, il, ilz)
             ilz = idx (lz_lo, ig, ik, it, ie, is)
! <TT
             if (idx_local(g_lo,iglo)) &
                  nn_from(proc_id(lz_lo,ilz)) = nn_from(proc_id(lz_lo,ilz)) + 1
             if (idx_local(lz_lo,ilz)) &
                  nn_to(proc_id(g_lo,iglo)) = nn_to(proc_id(g_lo,iglo)) + 1
          end do
       end do
    end do

    do ip = 0, nproc-1
       if (nn_from(ip) > 0) then
          allocate (from_list(ip)%first(nn_from(ip)))
          allocate (from_list(ip)%second(nn_from(ip)))
          allocate (from_list(ip)%third(nn_from(ip)))
       end if
       if (nn_to(ip) > 0) then
          allocate (to_list(ip)%first(nn_to(ip)))
          allocate (to_list(ip)%second(nn_to(ip)))
       end if
    end do

    ! get local indices of elements distributed to/from other processors
    nn_to = 0
    nn_from = 0
    do iglo = g_lo%llim_world, g_lo%ulim_world
! TT> expansion of gidx2lzidx
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       il = il_idx(g_lo,iglo)
! <TT
       do isign = 1, 2
! TT>
          if (isign==2) il = 2*g_lo%nlambda+1 - il
! <TT
          do ig = -ntgrid, ntgrid
! TT>
!             call gidx2lzidx (ig, isign, g_lo, iglo, lz_lo, ntgrid, il, ilz)
!             write(*,*) ig,':',isign,':',iglo,':',il,':',ilz
             ilz = idx (lz_lo, ig, ik, it, ie, is)
! <TT
             if (idx_local(g_lo,iglo)) then
                ip = proc_id(lz_lo,ilz)
                n = nn_from(ip) + 1
                nn_from(ip) = n
                from_list(ip)%first(n) = ig
                from_list(ip)%second(n) = isign
                from_list(ip)%third(n) = iglo
             end if
             if (idx_local(lz_lo,ilz)) then
                ip = proc_id(g_lo,iglo)
                n = nn_to(ip) + 1
                nn_to(ip) = n
                to_list(ip)%first(n) = il
                to_list(ip)%second(n) = ilz
             end if
          end do
       end do
    end do

    from_low (1) = -ntgrid
    from_low (2) = 1
    from_low (3) = g_lo%llim_proc

    to_low = lz_lo%llim_proc

    to_high(1) = 2*nlambda+1
    to_high(2) = lz_lo%ulim_alloc

    from_high(1) = ntgrid
    from_high(2) = 2
    from_high(3) = g_lo%ulim_alloc

    call init_redist (lorentz_map, 'c', to_low, to_high, to_list, &
         from_low, from_high, from_list)

    call delete_list (to_list)
    call delete_list (from_list)

    done = .true.

  end subroutine init_lorentz_redistribute

  subroutine solfp1 (g, gold, g1, phi, bpar, diagnostics)
    use agk_layouts, only: g_lo
    use theta_grid, only: ntgrid
    use run_parameters, only: fphi, fbpar
    use le_grids, only: integrate_moment
    use agk_time, only: dtime
    use dist_fn_arrays, only: c_rate
    use constants
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex,mem_dealloc
    implicit none
    complex, dimension (-ntgrid:,:,:), intent (in out) :: g, gold, g1 
    complex, dimension (-ntgrid:,:,:), intent (in) :: phi, bpar
    complex, dimension (:,:,:), allocatable :: gc1, gc2, gc3
    integer, optional, intent (in) :: diagnostics

    if (collision_model_switch == collision_model_none) return

    if (heating .and. present(diagnostics)) then
       call attempt_alloc("gc1",tcomplex,3,2*ntgrid+1,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
       allocate (gc1(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc)) ; gc1 = 0.
       call complete_alloc("gc1")
       call attempt_alloc("gc2",tcomplex,3,2*ntgrid+1,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
       allocate (gc2(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc)) ; gc2 = 0.
       call complete_alloc("gc2")
       call attempt_alloc("gc3",tcomplex,3,2*ntgrid+1,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
       allocate (gc3(-ntgrid:ntgrid,2,g_lo%llim_proc:g_lo%ulim_alloc)) ; gc3 = 0.
       call complete_alloc("gc3")
    else
       allocate (gc1(1,1,1)) ; gc1 = 0.
       allocate (gc2(1,1,1)) ; gc2 = 0.
       allocate (gc3(1,1,1)) ; gc3 = 0.
    end if
    
    if (adjust) call g_adjust (g, phi, bpar, fphi, fbpar)
    if (adjust) call g_adjust (gold, phi, bpar, fphi, fbpar)

! TT>
!!$! TT>
!!$!    if (present(diagnostics)) then
!!$    if (heating .and. present(diagnostics)) then
!!$! <TT
!!$       gc3 = g
!!$       call solfp_lorentz (g, gc1, gc2, diagnostics)
!!$    else
!!$       call solfp_lorentz (g, gc1, gc2)
!!$    end if
!!$    
!!$    if (conserve_momentum) call conserve_mom (g, g1)
    select case (collision_model_switch)

    case (collision_model_full)

       if (heating .and. present(diagnostics)) then
          gc3 = g
          call solfp_lorentz (g, gc1, gc2, diagnostics)
       else
          call solfp_lorentz (g, gc1, gc2)
       end if
       if (conserve_moments) call conserve_lorentz (g, g1)
!       if (conserve_momentum) call conserve_mom (g, g1)

       call solfp_ediffuse (g)
       if (conserve_moments) call conserve_diffuse (g, g1)

    case (collision_model_lorentz,collision_model_lorentz_test)

       if (heating .and. present(diagnostics)) then
          gc3 = g
          call solfp_lorentz (g, gc1, gc2, diagnostics)
       else
          call solfp_lorentz (g, gc1, gc2)
       end if
       if (conserve_moments) call conserve_lorentz (g, g1)
!       if (conserve_momentum) call conserve_mom (g, g1)

    case (collision_model_ediffuse)

       call solfp_ediffuse (g)
       if (conserve_moments) call conserve_diffuse (g, g1)

    end select
! <TT

    if (heating .and. present(diagnostics)) then

       call integrate_moment (gc1, c_rate(:,:,:,:,1))

       if (hyper_colls) call integrate_moment (gc2, c_rate(:,:,:,:,2))

! form (h_i+1 + h_i)/2 * C(h_i+1) and integrate.  

       gc3 = 0.5*conjg(g+gold)*(g-gc3)/dtime

      call integrate_moment (gc3, c_rate(:,:,:,:,3))

    end if

    if (heating .and. present(diagnostics)) then
       call mem_dealloc("gc1",tcomplex,3,2*ntgrid+1,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
       call mem_dealloc("gc2",tcomplex,3,2*ntgrid+1,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
       call mem_dealloc("gc3",tcomplex,3,2*ntgrid+1,2,g_lo%ulim_alloc-g_lo%llim_proc+1)
    endif
    deallocate (gc1, gc2, gc3)
    
    if (adjust) call g_adjust (g, phi, bpar, -fphi, -fbpar)
    if (adjust) call g_adjust (gold, phi, bpar, -fphi, -fbpar)
    
  end subroutine solfp1

! TT>
!  subroutine conserve_mom (g, g1)
  subroutine conserve_lorentz (g, g1)
! <TT

    use mp, only: proc0
    use theta_grid, only: ntgrid
    use species, only: nspec
    use kgrids, only: naky, nakx
    use agk_layouts, only: g_lo, ik_idx, it_idx, ie_idx, is_idx
    use le_grids, only: integrate_moment
    use dist_fn_arrays, only: aj0, aj1vp2, vpa
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex,mem_dealloc
    implicit none
    
    complex, dimension (-ntgrid:,:,g_lo%llim_proc:), intent (in out) :: g, g1
    complex, dimension (:,:,:,:), allocatable :: v0y0, v1y0

    integer :: ig, isgn, iglo, ik, ie, il, is, it, all = 1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! First get v0y0

    call attempt_alloc("v0y0",tcomplex,4,2*ntgrid+1,nakx, naky, nspec)
    allocate (v0y0(-ntgrid:ntgrid, nakx, naky, nspec))         
    call complete_alloc("v0y0")

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       g1(:,:,iglo) = 0.5 * vnew_D(ie,is)*aj1vp2(iglo) * g(:,:,iglo)
    end do

    call integrate_moment (g1, v0y0, all)    ! v0y0
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get v1y0

    call attempt_alloc("v1y0",tcomplex,4,2*ntgrid+1,nakx, naky, nspec)
    allocate (v1y0(-ntgrid:ntgrid, nakx, naky, nspec))         
    call complete_alloc("v1y0")

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       do isgn = 1, 2
          g1(:,isgn,iglo) = vnew_D(ie,is)*vpa(isgn,iglo)*aj0(iglo) &
               * g(:,isgn,iglo)
       end do
    end do

    call integrate_moment (g1, v1y0, all)    ! v1y0

!    if (proc0) then
!       write (*,*) v1y0
!    end if

! Conserve momentum:

!    write (*,*) sum(z0)
    
    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       do isgn=1,2
! TT> changed sign of z0
!          g(:,isgn,iglo) = g(:,isgn,iglo) + z0(isgn,iglo)*v0y0(:,it,ik,is) &
!               - z1(isgn,iglo) * v1y0(:,it,ik,is)
          g(:,isgn,iglo) = g(:,isgn,iglo) - z0(iglo)*v0y0(:,it,ik,is) &
               - z1(isgn,iglo) * v1y0(:,it,ik,is)
! <TT
       end do
    end do

    call mem_dealloc("v0y0",tcomplex,4,2*ntgrid+1,nakx, naky, nspec)
    call mem_dealloc("v1y0",tcomplex,4,2*ntgrid+1,nakx, naky, nspec)
    deallocate (v0y0, v1y0)

! TT>
!  end subroutine conserve_mom
  end subroutine conserve_lorentz
! <TT

! TT>
  subroutine conserve_diffuse (g, g1)

    use mp, only: proc0
    use theta_grid, only: ntgrid
    use species, only: nspec
    use kgrids, only: naky, nakx
    use agk_layouts, only: g_lo, ik_idx, it_idx, ie_idx, is_idx
    use le_grids, only: e, al, integrate_moment, negrid
    use dist_fn_arrays, only: aj0, aj1vp2, vpa

    complex, dimension (-ntgrid:,:,g_lo%llim_proc:), intent (in out) :: g, g1
    complex, dimension (:,:,:,:), allocatable :: v0y0, v1y0, v2y0
! TT>
!    real, dimension (:,:), allocatable :: vns
!    real :: fac
! <TT

    integer :: ig, isgn, iglo, ik, ie, il, is, it, all = 1

! TT>
!    allocate (vns(negrid,nspec))
!    vns = vnew_s - vnew_D
! <TT

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! First get v0y0

    allocate (v0y0(-ntgrid:ntgrid, nakx, naky, nspec))         

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT>
!       fac = 0.5 * (vnew_s(ie,is) - vnew_D(ie,is)) * aj1vp2(iglo)
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! <TT
! V_perp == e(ie,is)*al(il)*aj1(ig,iglo)
! v0 = nu_s V_perp
! no factor of sqrt( kperp2 smz**2 / bmag ) in v0 because it was absorbed into u0 -- MAB
! TT>
!             g1(ig,isgn,iglo) = vns(ik,ie,is)*e(ie,is)*al(il)*aj1(ig,iglo) &
!                  * g(ig,isgn,iglo)
!             g1(ig,isgn,iglo) = 0.5*vns(ie,is)*aj1vp2(iglo) * g(ig,isgn,iglo)
!             g1(ig,isgn,iglo) = fac * g(ig,isgn,iglo)
!          end do
!       end do
       g1(:,:,iglo) = 0.5 * (vnew_s(ie,is) - vnew_D(ie,is)) * aj1vp2(iglo) * g(:,:,iglo)
! <TT
    end do

    call integrate_moment (g1, v0y0, all)    ! v0y0
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Now get v1y0 and v2y0 (actually v1 and v2 . (y0 - v0y0 * s0)...i.e. v1y1 and v2y1)

    allocate (v1y0(-ntgrid:ntgrid, nakx, naky, nspec))         

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       it = it_idx(g_lo,iglo)
       ik = ik_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT>
!       do isgn = 1, 2
!          fac = (vnew_s(ie,is) - vnew_D(ie,is)) * vpa(isgn,iglo) * aj0(iglo)
!          do ig=-ntgrid, ntgrid
! <TT
! v_parallel == vpa
! V_parallel == v_parallel J0
! v1 = nu V_parallel f_0
!
! TT>
!             g1(ig,isgn,iglo) = vns(ik,ie,is)*vpa(ig,isgn,iglo)*aj0(ig,iglo) &
!                  * (g(ig,isgn,iglo) - bs0(ig,isgn,iglo)*v0y0(ig,it,ik,is))
!             g1(ig,isgn,iglo) = vns(ie,is)*vpa(isgn,iglo)*aj0(iglo) &
!                  * (g(ig,isgn,iglo) - bs0(iglo)*v0y0(ig,it,ik,is))
!             g1(ig,isgn,iglo) = fac * (g(ig,isgn,iglo) - bs0(iglo)*v0y0(ig,it,ik,is))
!          end do
!       end do
       g1 (:,:,iglo) = (vnew_s(ie,is)-vnew_D(ie,is)) * spread(vpa(:,iglo), 1, ntgrid*2+1) * aj0(iglo) &
            * (g(:,:,iglo) - bs0(iglo)*spread(v0y0(:,it,ik,is),2,2))
! <TT
    end do

    call integrate_moment (g1, v1y0, all)    ! v1y1

    allocate (v2y0(-ntgrid:ntgrid, nakx, naky, nspec))         

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       it = it_idx(g_lo,iglo)
       ik = ik_idx(g_lo,iglo)
       ie = ie_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
! TT>
!       fac = vnew_E(ie,is) * e(ie,is) * aj0(iglo)
!       do isgn = 1, 2
!          do ig=-ntgrid, ntgrid
! <TT
! v2 = nu_E e J0
! TT>
!             g1(ig,isgn,iglo) = vnmult(2)*vnew_E(ik,ie,is)*e(ie,is)*aj0(ig,iglo) &
!                  * (g(ig,isgn,iglo) - bs0(ig,isgn,iglo)*v0y0(ig,it,ik,is))
!             g1(ig,isgn,iglo) = vnew_E(ie,is)*e(ie,is)*aj0(iglo) &
!                  * (g(ig,isgn,iglo) - bs0(iglo)*v0y0(ig,it,ik,is))
!             g1(ig,isgn,iglo) = fac * (g(ig,isgn,iglo) - bs0(iglo)*v0y0(ig,it,ik,is))
!          end do
!       end do
       g1(:,:,iglo) = vnew_E(ie,is) * e(ie,is) * aj0(iglo) &
            & * (g(:,:,iglo) - bs0(iglo)*spread(v0y0(:,it,ik,is),2,2))
! <TT
    end do

    call integrate_moment (g1, v2y0, all)    ! v2y1

! TT>
!    deallocate (vns)
! <TT

! Conserve momentum and energy:

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
       is = is_idx(g_lo,iglo)
       do isgn=1,2
! TT>
!          do ig=-ntgrid,ntgrid
! TT: reduced dimension of bs0, bw0 & bz0
!             g(ig,isgn,iglo) = g(ig,isgn,iglo) - bs0(ig,isgn,iglo)*v0y0(ig,it,ik,is) &
!                  - bz0(ig,isgn,iglo)*v2y0(ig,it,ik,is) + (bz0(ig,isgn,iglo)*v2w1(ig,it,ik,is) &
!                  - bw0(ig,isgn,iglo))*v1y0(ig,it,ik,is)
!             g(ig,isgn,iglo) = g(ig,isgn,iglo) - bs0(iglo)*v0y0(ig,it,ik,is) &
!                  - bz0(isgn,iglo)*v2y0(ig,it,ik,is) + (bz0(isgn,iglo)*v2w1(it,ik,is) &
!                  - bw0(isgn,iglo))*v1y0(ig,it,ik,is)
!          end do
          g(:,isgn,iglo) = g(:,isgn,iglo) - bs0(iglo)*v0y0(:,it,ik,is) &
               - bz0(isgn,iglo)*v2y0(:,it,ik,is) + (bz0(isgn,iglo)*v2w1(it,ik,is) &
               - bw0(isgn,iglo))*v1y0(:,it,ik,is)
! <TT
       end do
    end do

    deallocate (v0y0, v1y0, v2y0)

  end subroutine conserve_diffuse
! <TT

  subroutine g_adjust (g, phi, bpar, facphi, facbpar)
    use species, only: spec
    use theta_grid, only: ntgrid
! TT> removed anon
!    use le_grids, only: anon
! <TT
    use dist_fn_arrays, only: aj0, aj1vp2
    use agk_layouts, only: g_lo, ik_idx, it_idx, is_idx
    implicit none
    complex, dimension (-ntgrid:,:,g_lo%llim_proc:), intent (in out) :: g
    complex, dimension (-ntgrid:,:,:), intent (in) :: phi, bpar
    real, intent (in) :: facphi, facbpar

    integer :: iglo, ig, ik, it, is
    complex :: adj

    do iglo = g_lo%llim_proc, g_lo%ulim_proc
       ik = ik_idx(g_lo,iglo)
       it = it_idx(g_lo,iglo)
! TT>
!       ie = ie_idx(g_lo,iglo)
! <TT
       is = is_idx(g_lo,iglo)
       do ig = -ntgrid, ntgrid
! TT> removed anon
!          adj = anon(ie,is)*aj1vp2(iglo)*bpar(ig,it,ik)*facbpar &
!               + spec(is)%z*anon(ie,is)*phi(ig,it,ik)*aj0(iglo) &
          adj = aj1vp2(iglo)*bpar(ig,it,ik)*facbpar &
               + spec(is)%z*phi(ig,it,ik)*aj0(iglo) &
! <TT
                  /spec(is)%temp*facphi
          g(ig,1,iglo) = g(ig,1,iglo) + adj
          g(ig,2,iglo) = g(ig,2,iglo) + adj
       end do
    end do
  end subroutine g_adjust

  subroutine solfp_lorentz (g, gc, gh, diagnostics)
    use species, only: spec, electron_species
    use theta_grid, only: ntgrid
    use le_grids, only: nlambda
    use agk_layouts, only: g_lo, lz_lo, is_idx
    use redistribute, only: gather, scatter
    implicit none
    complex, dimension (-ntgrid:,:,g_lo%llim_proc:), intent (in out) :: g, gc, gh
    integer, optional, intent (in) :: diagnostics

    complex, dimension (2*nlambda+1) :: delta
    complex :: fac
    integer :: iglo, ilz, il, is, je

    call gather (lorentz_map, g, glz)

    if (heating .and. present(diagnostics)) then
       do ilz = lz_lo%llim_proc, lz_lo%ulim_proc

          je = 2*nlambda

          do il = 1, je-1
             fac = glz(il+1,ilz)-glz(il,ilz)
             glzc(il,ilz) = conjg(fac)*fac*d1(il,ilz)  ! d1 accounts for hC(h) entropy
          end do
       end do
       call scatter (lorentz_map, glzc, gc)

       if (hyper_colls) then
          do ilz = lz_lo%llim_proc, lz_lo%ulim_proc
             
             je = 2*nlambda 
             
             do il = 1, je-1
                fac = glz(il+1,ilz)-glz(il,ilz)
                glzc(il,ilz) = conjg(fac)*fac*h1(il,ilz)  ! h1 accounts for hH(h) entropy
             end do
          end do
          call scatter (lorentz_map, glzc, gh)
       end if
    end if

    ! solve for glz row by row
    do ilz = lz_lo%llim_proc, lz_lo%ulim_proc
       is = is_idx(lz_lo,ilz)
! TT>
!       if (abs(vnew(ik,1,is)) < 2.0*epsilon(0.0)) cycle
       if (abs(vnew(1,is)) < 2.0*epsilon(0.0)) cycle
! <TT

       je = 2*nlambda+1

       glz(je:,ilz) = 0.0

       ! right and left sweeps for tridiagonal solve:

       delta(1) = glz(1,ilz)
       do il = 1, je-1
          delta(il+1) = glz(il+1,ilz) - ql(il+1,ilz)*delta(il)
       end do
       
       glz(je,ilz) = delta(je)*betaa(je,ilz)
       do il = je-1, 1, -1
          glz(il,ilz) = (delta(il) - c1(il,ilz)*glz(il+1,ilz))*betaa(il,ilz)
       end do

    end do

    call scatter (lorentz_map, glz, g)

  end subroutine solfp_lorentz

! TT>
  subroutine solfp_ediffuse (g)
    use species, only: spec
    use theta_grid, only: ntgrid
    use le_grids, only: negrid
    use redistribute, only: gather, scatter
    use agk_layouts, only: is_idx, e_lo, g_lo

    implicit none

    complex, dimension (-ntgrid:,:,g_lo%llim_proc:), intent (in out) :: g
    complex, dimension (negrid) :: delta
    integer :: ielo, ie, is

    call gather (ediffuse_map, g, ged)

    ! solve for ged row by row
    do ielo = e_lo%llim_proc, e_lo%ulim_proc
       is = is_idx(e_lo,ielo)

       if (spec(is)%nu < 2.0*epsilon(0.0)) cycle

       delta(1) = ged(1,ielo)
! should be ie = 1, negrid? and pad ged, eql, etc. with extra element?
       do ie = 1, negrid-1
          delta(ie+1) = ged(ie+1,ielo) - eql(ie+1,ielo)*delta(ie)
       end do
       
! also following should change to mimic lorentz?
       ged(negrid+1,ielo) = 0.0
       do ie = negrid, 1, -1
          ged(ie,ielo) = (delta(ie) - ec1(ie,ielo)*ged(ie+1,ielo))*ebetaa(ie,ielo)
       end do

    end do

    call scatter (ediffuse_map, ged, g)

  end subroutine solfp_ediffuse
! <TT

  subroutine reset_init
!
! forces recalculation of coefficients in collision operator
! when timestep changes.
!    
    initialized = .false.  

  end subroutine reset_init

end module collisions
