module fields
  use fields_arrays, only: phi, apar, bpar, phinew, aparnew, bparnew
  use fields_arrays, only: phi_ext, apar_ext

  implicit none

  public :: init_fields  ! initialize this module
  public :: advance      ! do a full time step
  public :: phinorm      ! diagnostic variable of no intrinsic value
! TT> They should be used from fields_arrays
!  public :: phi, apar, bpar, phinew, aparnew, bparnew  ! field variables at old and new timestep
! <TT
  public :: reset_init   ! Sets fields to zero.  Useful for certain kinds of restarts.

  private

  logical :: initialized = .false.
  logical :: debug=.false.

contains

  subroutine init_fields
    use mp, only: proc0
    use theta_grid, only: init_theta_grid
    use run_parameters, only: init_run_parameters
!    use dist_fn, only: init_dist_fn, ginit
!    use init_g, only: init_init_g
    use dist_fn, only: init_dist_fn
    use init_g, only: init_init_g, ginit
    use fields_implicit, only: init_fields_implicit, init_phi_implicit
    use nonlinear_terms, only: nl_finish_init => finish_init
    use antenna, only: init_antenna
    implicit none
    logical :: restarted

    if (initialized) return  ;  initialized = .true.

    call init_theta_grid       ! Set up grid along the field line.
    if (debug.and.proc0) write(*,*)'Theta Grid initialized'   !DEBUG-GGH-080324
    call init_init_g           ! Initialize module that handles initial conditions
    if (debug.and.proc0) write(*,*)'init_g initialized'   !DEBUG-GGH-080324
    call init_run_parameters   ! Get basic inputs, set up perpendicular grids.
    if (debug.and.proc0) write(*,*)'Run Parameters initialized'   !DEBUG-GGH-080324
    call init_dist_fn          ! Make sure code is ready to calculate with distribution function
    if (debug.and.proc0) write(*,*)'Dist Func initialized'   !DEBUG-GGH-080324
    call allocate_arrays       ! Allocate memory as necessary
    if (debug.and.proc0) write(*,*)'Arrays allocated'   !DEBUG-GGH-080324
    call init_fields_implicit  ! Calculate matrices for implicit soln of Maxwell eqns.
    if (debug.and.proc0) write(*,*)'Implicit Fields initialized'   !DEBUG-GGH-080324

! Turn on nonlinear terms if necessary.
    call nl_finish_init
    if (debug.and.proc0) write(*,*)'NL Finish initialized'   !DEBUG-GGH-080324

    call ginit (restarted)
    if (debug.and.proc0) write(*,*)'Ginit initialized'   !DEBUG-GGH-080324
    call init_antenna
    if (debug.and.proc0) write(*,*)'Antenna initialized'   !DEBUG-GGH-080324

    if (restarted) return

    call init_phi_implicit  ! actually: set up fields for first time step.
    if (debug.and.proc0) write(*,*)'Phi Implicit initialized'   !DEBUG-GGH-080324
    
  end subroutine init_fields

  subroutine allocate_arrays
    use theta_grid, only: ntgrid
    use kgrids, only: naky, nakx
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex
    implicit none
    logical :: alloc = .true.

    if (alloc) then
       call attempt_alloc("phi",tcomplex,3,2*ntgrid+1,nakx,naky)
       allocate (     phi (-ntgrid:ntgrid,nakx,naky))
       call complete_alloc("phi")
       call attempt_alloc("apar",tcomplex,3,2*ntgrid+1,nakx,naky)
       allocate (    apar (-ntgrid:ntgrid,nakx,naky))
       call complete_alloc("apar")
       call attempt_alloc("bpar",tcomplex,3,2*ntgrid+1,nakx,naky)
       allocate (    bpar (-ntgrid:ntgrid,nakx,naky))
       call complete_alloc("bpar")
       call attempt_alloc("phinew",tcomplex,3,2*ntgrid+1,nakx,naky)
       allocate (  phinew (-ntgrid:ntgrid,nakx,naky))
       call complete_alloc("phinew")
       call attempt_alloc("aparnew",tcomplex,3,2*ntgrid+1,nakx,naky)
       allocate ( aparnew (-ntgrid:ntgrid,nakx,naky))
       call complete_alloc("aparnew")
       call attempt_alloc("bparnew",tcomplex,3,2*ntgrid+1,nakx,naky)
       allocate (bparnew  (-ntgrid:ntgrid,nakx,naky))
       call complete_alloc("bparnew")
       call attempt_alloc("apar_ext",tcomplex,3,2*ntgrid+1,nakx,naky)
       allocate (apar_ext (-ntgrid:ntgrid,nakx,naky))
       call complete_alloc("apar_ext")
    endif

    phi  = 0. ; phinew  = 0.
    apar = 0. ; aparnew = 0.
    bpar = 0. ; bparnew = 0.

    apar_ext = 0.

    alloc = .false.
  end subroutine allocate_arrays

! TT>
!  subroutine advance (istep, aant)
  subroutine advance (istep)
! <TT
    use fields_implicit, only: advance_implicit
    implicit none
    integer, intent (in) :: istep
! TT>
!    logical, intent (in) :: aant
! <TT

! Could call different algorithm here...e.g., something explicit.

! TT>
!    call advance_implicit (istep, aant)
    call advance_implicit (istep)
! <TT

  end subroutine advance

  subroutine phinorm (phitot)
! TT>
!    use theta_grid, only: delthet
    use theta_grid, only: delthet, ntgrid
! <TT
    use kgrids, only: naky, nakx
    use constants
    implicit none
    real, dimension (:,:), intent (out) :: phitot
    integer :: ik, it

! GS2 reported the square of this value.  Only used for rough diagnostic purposes.

    ! TT: ERROR: This is double-counting the end point for per BC
    ! We should sum (-ntgrid:ntgrid-1) :TT
    do ik = 1, naky
       do it = 1, nakx
          phitot(it,ik) = 0.5/pi &
! TT>
!           *(sum((cabs(phinew(:,it,ik)) + cabs(aparnew(:,it,ik)) &
!                  + cabs(bparnew(:,it,ik)))*delthet))
           *(sum((cabs(phinew(:ntgrid-1,it,ik)) + cabs(aparnew(:ntgrid-1,it,ik)) &
                  + cabs(bparnew(:ntgrid-1,it,ik)))*delthet))
! <TT
       end do
    end do

  end subroutine phinorm

  subroutine reset_init

    initialized  = .false.
    phi  = 0.  ;  phinew  = 0.
    apar = 0.  ;  aparnew = 0.  ! GS2 did not have these statements!
    bpar = 0.  ;  bparnew = 0.  ! GS2 did not have these statements!

  end subroutine reset_init

end module fields
