!
! This module is much simpler in AstroGK than GS2
! because of the lack of equilibrium magnetic shear. 
! Shear couples different modes together linearly
! through the twist and shift boundary conditions, 
! and leads to very large matrix inversions which 
! are parallelized in GS2.  
!
! The only routine which would benefit strongly from 
! parallelization in the AstroGK case is getfield.
! Getfield is called in the timestepping loop.  
! (Note that advance_implicit is already parallel.)
! Getfield calls a couple of array-packing and 
! unpacking routines, which could be made parallel
! if getfield were parallelized.  
!
! One obvious strategy for parallelizing these loops 
! would be to distribute the it and ik indices.  Since
! many large runs use layouts with these indices on 
! processor (to optimize the evaluation of the nonlinear
! terms), I am putting this off for another time.
!

module fields_implicit
  use fields_arrays, only: nidx
  implicit none

  public :: init_fields_implicit   ! initialize this module
  public :: advance_implicit       ! Do a full timestep using original algorithm
  public :: init_phi_implicit      ! Set up fields for first time step
  public :: nidx                   ! Do not recall why this isn't declared locally to fields_implicit module.
  public :: reset_init             ! Deallocates matrices for implicit fields solution, in prep for changing dt.

  private

  integer, save :: nfield          ! number of fields (1,2, or 3)
  logical :: initialized = .false.

contains

  subroutine init_fields_implicit

    use antenna, only: init_antenna
    use theta_grid, only: init_theta_grid
    use kgrids, only: init_kgrids
    implicit none

    if (initialized) return
    initialized = .true.

    call init_theta_grid          ! Set up grid along the field line
    call init_kgrids              ! Set up grids perpendicular to the field line
    call init_response_matrix     ! Calculate matrices for implicit Maxwell solve
    call init_antenna             ! Set up external driving antenna

  end subroutine init_fields_implicit

  subroutine init_phi_implicit
    use fields_arrays, only: phi, apar, bpar, phinew, aparnew, bparnew
    use dist_fn, only: get_init_field
    use dist_fn_arrays, only: g, gnew
    implicit none

    call init_fields_implicit
! TT> next line is not good for the initialization of the EM problem
! TT: I'll leave it here in case we want to compare with GS2 result
!    call getfield (phinew, aparnew, bparnew)
!    phi = phinew; apar = aparnew; bpar = bparnew
    ! bug fix follows
    call get_init_field (phinew, aparnew, bparnew)
    ! next line does not hurt, but maybe not needed except for heating diag
    phi = phinew; apar = aparnew; bpar = bparnew; g = gnew
! <TT

  end subroutine init_phi_implicit

! TT>
!  subroutine advance_implicit (istep, aant)
  subroutine advance_implicit (istep)
! <TT
    use fields_arrays, only: phi, apar, bpar, phinew, aparnew, bparnew
    use fields_arrays, only: apar_ext 
    use antenna, only: antenna_amplitudes
    use dist_fn, only: timeadv, exb_shear
    use dist_fn_arrays, only: g, gnew
! TT>
    use run_parameters, only: aant
! <TT
    implicit none
    integer, intent (in) :: istep
! TT>
!    logical, intent (in) :: aant
! <TT
    
    call antenna_amplitudes (apar_ext)                                          ! GGH NOTE: apar_ext is initialized in this call
    call exb_shear (gnew, phinew, aparnew, bparnew)                             ! See Hammett & Loureiro, APS 2006
    g = gnew  ;  phi = phinew  ;  apar = aparnew  ;  bpar = bparnew             ! Store previously advanced values as current values
    call timeadv (phi, apar, bpar, phinew, aparnew, bparnew, istep)             ! Advance f
    if (aant) then
       aparnew = aparnew + apar_ext                                                ! Add in antenna
    else
       phinew = phinew + apar_ext
    end if
    call getfield (phinew, aparnew, bparnew)                                    ! Find future part of fields
    phinew = phinew + phi ; aparnew = aparnew + apar ; bparnew = bparnew + bpar ! Update fields
    call timeadv (phi, apar, bpar, phinew, aparnew, bparnew, istep)             ! Complete f advance with updated fields

  end subroutine advance_implicit

  subroutine getfield (phi, apar, bpar)
! TT>
    use mp, only: nproc, sum_allreduce
! <TT
    use kgrids, only: naky, nakx
    use fields_arrays, only: aminv
    use theta_grid, only: ntgrid
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex,mem_dealloc
    use agk_layouts, only: f_lo, ik_idx, it_idx
    implicit none
    complex, dimension (-ntgrid:,:,:), intent (in) :: phi, apar, bpar
    complex, dimension (:,:,:), allocatable :: fl
    complex, dimension (:), allocatable :: u
    integer :: ik, it, i, iflo

    call attempt_alloc("fl", tcomplex, 3, nidx, nakx, naky)
    allocate (fl(nidx, nakx, naky))
    call complete_alloc("fl")

! BD: Double-check the last argument:
! TT>
!!$    call attempt_alloc("u", tcomplex, 1, f_lo%ulim_alloc-f_lo%llim_proc+1)  
!!$    allocate (u(f_lo%llim_proc:f_lo%ulim_alloc)) ; u = 0.
    call attempt_alloc("u", tcomplex, 1, f_lo%ulim_world-f_lo%llim_world+1)  
    allocate (u(f_lo%llim_world:f_lo%ulim_world)) ; u = 0.
! <TT
    call complete_alloc("u")

    ! am*u = fl, Poisson's and Ampere's law, u is phi, apar, bpar 
    ! u = aminv*fl

    call get_field_vector (fl, phi, apar, bpar)
    
    do iflo = f_lo%llim_proc, f_lo%ulim_proc
       u(iflo) = - sum(aminv(:,iflo) * fl(:, it_idx(f_lo, iflo), ik_idx(f_lo, iflo)))
    end do

    call mem_dealloc("fl",tcomplex,3, nidx, nakx, naky)
    deallocate (fl)

! TT> This should be the minimal communication
    if (nproc > 1) call sum_allreduce (u)
! <TT

    call get_field_solution (u)

! BD: Double-check the last argument:
! TT>
!!$    call mem_dealloc("u", tcomplex, 1, f_lo%ulim_alloc-f_lo%llim_proc+1)
    call mem_dealloc("u", tcomplex, 1, f_lo%ulim_world-f_lo%llim_world+1)
! <TT
    deallocate (u)

  end subroutine getfield

  subroutine get_field_vector (fl, phi, apar, bpar)
    use theta_grid, only: ntgrid
    use kgrids, only: naky, nakx
    use dist_fn, only: getfieldeq
    use run_parameters, only: use_Phi, use_Apar, use_Bpar
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex,mem_dealloc
    implicit none
    complex, dimension (-ntgrid:,:,:), intent (in) :: phi, apar, bpar
    complex, dimension (:,:,:), intent (out) :: fl
    complex, dimension (:,:,:), allocatable :: fieldeq, fieldeqa, fieldeqp
    integer :: istart

    call attempt_alloc("fieldeq",tcomplex,3, 2*ntgrid+1, nakx, naky)
    allocate (fieldeq (-ntgrid:ntgrid,nakx,naky))
    call complete_alloc("fieldeq")
    call attempt_alloc("fieldeqa",tcomplex,3, 2*ntgrid+1, nakx, naky)
    allocate (fieldeqa(-ntgrid:ntgrid,nakx,naky))
    call complete_alloc("fieldeqa")
    call attempt_alloc("fieldeqp",tcomplex,3, 2*ntgrid+1, nakx, naky)
    allocate (fieldeqp(-ntgrid:ntgrid,nakx,naky))
    call complete_alloc("fieldeqp")

    call getfieldeq (phi, apar, bpar, fieldeq, fieldeqa, fieldeqp)

    istart = 0

    if (use_Phi) then
       istart = istart + 1
       fl(istart:nidx:nfield,:,:) = fieldeq   ! BD Coding is different from GS2
    end if

    if (use_Apar) then
       istart = istart + 1
       fl(istart:nidx:nfield,:,:) = fieldeqa
    end if

    if (use_Bpar) then
       istart = istart + 1
       fl(istart:nidx:nfield,:,:) = fieldeqp
    end if

    call mem_dealloc("fieldeq",tcomplex,3, 2*ntgrid+1, nakx, naky)
    call mem_dealloc("fieldeqa",tcomplex,3, 2*ntgrid+1, nakx, naky)
    call mem_dealloc("fieldeqp",tcomplex,3, 2*ntgrid+1, nakx, naky)
    deallocate (fieldeq, fieldeqa, fieldeqp)

  end subroutine get_field_vector

  subroutine get_field_solution (u)
    use theta_grid, only: ntgrid
    use fields_arrays, only: phinew, aparnew, bparnew
    use run_parameters, only: use_Phi, use_Apar, use_Bpar
! TT>
!!$    use agk_layouts, only: f_lo!, if_idx, ik_idx, it_idx
!!$    use agk_layouts, only: idx_local, proc_id
!!$    use mp, only: nproc
    use agk_layouts, only: f_lo
    use kgrids, only: nakx, naky
! <TT
    implicit none
! TT>
!!$    complex, dimension (f_lo%llim_proc:), intent (in) :: u
!!$    integer :: istart, iflo, iff, ik, it, ig
    complex, dimension (f_lo%llim_world:), intent (in) :: u
    integer :: istart, iflo, ik, it, ig
! <TT

    istart = 0

! BD Could simplify these loops?  Look at current GS2 coding.
! TT: I think I've simplified it.  Which one do you like?
    if (use_Phi) then
       istart = istart + 1
! TT>
!!$       do iflo = f_lo%llim_world, f_lo%ulim_world
!!$          iff = if_idx(f_lo,iflo)
!!$          if (mod((iff - istart),nfield) /= 0) cycle
!!$          ik = ik_idx(f_lo,iflo)
!!$          it = it_idx(f_lo,iflo)
!!$          ig = -ntgrid + (iff-istart)/nfield
!!$          if (idx_local(f_lo,iflo)) phinew(ig,it,ik) = u(iflo)
!!$          call broadcast (phinew(ig,it,ik), proc_id(f_lo,iflo))
!!$       end do
       ! TT: We could either do this:
!!$       phinew = reshape(u(istart-1::nfield), shape(phinew))
       ! TT: or this:
       iflo = istart - 1
       do ik=1, naky
          do it=1, nakx
             do ig=-ntgrid, ntgrid
                phinew(ig,it,ik) = u(iflo)
                iflo = iflo + nfield
             end do
          end do
       end do
! <TT
    endif

    if (use_Apar) then
       istart = istart + 1
! TT>
!!$       do iflo = f_lo%llim_world, f_lo%ulim_world
!!$          iff = if_idx(f_lo,iflo)
!!$          if (mod((iff - istart),nfield) /= 0) cycle
!!$          ik = ik_idx(f_lo,iflo)
!!$          it = it_idx(f_lo,iflo)
!!$          ig = -ntgrid + (iff-istart)/nfield
!!$          if (idx_local(f_lo,iflo)) aparnew(ig,it,ik) = u(iflo)
!!$          call broadcast (aparnew(ig,it,ik), proc_id(f_lo,iflo))
!!$       end do
       iflo = istart - 1
       do ik=1, naky
          do it=1, nakx
             do ig=-ntgrid, ntgrid
                aparnew(ig,it,ik) = u(iflo)
                iflo = iflo + nfield
             end do
          end do
       end do
! <TT
    endif

    if (use_Bpar) then
       istart = istart + 1
! TT>
!!$       do iflo = f_lo%llim_world, f_lo%ulim_world
!!$          iff = if_idx(f_lo,iflo)
!!$          if (mod((iff - istart),nfield) /= 0) cycle
!!$          ik = ik_idx(f_lo,iflo)
!!$          it = it_idx(f_lo,iflo)
!!$          ig = -ntgrid + (iff-istart)/nfield
!!$          if (idx_local(f_lo,iflo)) bparnew(ig,it,ik) = u(iflo)
!!$          call broadcast (bparnew(ig,it,ik), proc_id(f_lo,iflo))
!!$       end do
       iflo = istart - 1
       do ik=1, naky
          do it=1, nakx
             do ig=-ntgrid, ntgrid
                bparnew(ig,it,ik) = u(iflo)
                iflo = iflo + nfield
             end do
          end do
       end do
! <TT
   endif

  end subroutine get_field_solution

  subroutine reset_init
    use fields_arrays, only: aminv, nidx
    use kgrids, only: naky, nakx
    use agk_mem, only: tcomplex,mem_dealloc
    use agk_layouts, only: f_lo
    implicit none
    initialized = .false.
    
! BD Double check last argument:
    call mem_dealloc("aminv", tcomplex, 2, nidx, f_lo%ulim_alloc-f_lo%ulim_proc+1)
    deallocate (aminv)

  end subroutine reset_init

  subroutine init_response_matrix

    use fields_arrays, only: phi, apar, bpar, phinew, aparnew, bparnew
    use theta_grid, only: ntgrid,ntheta
    use kgrids, only: naky, nakx
    use dist_fn_arrays, only: g
    use run_parameters, only: use_Phi, use_Apar, use_Bpar
    use agk_layouts, only: init_fields_layouts, f_lo
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex,mem_dealloc
    use mp, only: proc0
    implicit none
    integer :: ig, ifield
    complex, dimension(:,:), allocatable :: am

    nfield = 0
    if (use_Phi)  nfield = nfield + 1
    if (use_Apar) nfield = nfield + 1
    if (use_BPar) nfield = nfield + 1
    nidx = (2*ntgrid+1)*nfield

    call init_fields_layouts (nidx, naky, nakx)

! BD Double check last argument:
! TT: I replaced ulim_alloc - ulim_proc -> ulim_alloc - llim_proc
    call attempt_alloc("am", tcomplex, 2, nidx, f_lo%ulim_alloc-f_lo%llim_proc+1)
    allocate (am(nidx, f_lo%llim_proc:f_lo%ulim_alloc))
    call complete_alloc("am")

    am = 0.0
    g = 0.0
    
    phi = 0.0
    apar = 0.0
    bpar = 0.0
    phinew = 0.0
    aparnew = 0.0
    bparnew = 0.0
    
    do ig = -ntgrid, ntgrid
       ifield = 0
       if (use_Phi) then
          ifield = ifield + 1
          phinew(ig,:,:) = 1.0
          call init_response_row (ig, ifield, am)
          phinew = 0.0
       end if
       
       if (use_Apar) then
          ifield = ifield + 1
          aparnew(ig,:,:) = 1.0
          call init_response_row (ig, ifield, am)
          aparnew = 0.0
       end if
       
       if (use_Bpar) then
          ifield = ifield + 1
          bparnew(ig,:,:) = 1.0
          call init_response_row (ig, ifield, am)
          bparnew = 0.0
       end if
    end do
    
    call init_inverse_matrix (am)
    call mem_dealloc("am",tcomplex,4,nidx, nidx, nakx, naky)
    deallocate (am)
    
  end subroutine init_response_matrix

  subroutine init_response_row (ig, ifield, am)

    use fields_arrays, only: phi, apar, bpar, phinew, aparnew, bparnew
    use theta_grid, only: ntgrid
    use kgrids, only: naky, nakx
    use dist_fn, only: getfieldeq, timeadv
    use run_parameters, only: use_Phi, use_Apar, use_Bpar
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex,mem_dealloc
    use agk_layouts, only: f_lo, idx, idx_local
    implicit none
    integer, intent (in) :: ig, ifield
    complex, dimension(:,f_lo%llim_proc:), intent (in out) :: am
    complex, dimension (:,:,:), allocatable :: fieldeq, fieldeqa, fieldeqp
    integer :: irow, istart, ik, it, ifin, iflo

    call attempt_alloc("fieldeq", tcomplex, 3, 2*ntgrid+1, nakx, naky)
    allocate (fieldeq (-ntgrid:ntgrid, nakx, naky))
    call complete_alloc("fieldeq")

    call attempt_alloc("fieldeqa", tcomplex, 3, 2*ntgrid+1, nakx, naky)
    allocate (fieldeqa(-ntgrid:ntgrid, nakx, naky))
    call complete_alloc("fieldeqa")

    call attempt_alloc("fieldeqp", tcomplex, 3, 2*ntgrid+1, nakx, naky)
    allocate (fieldeqp(-ntgrid:ntgrid, nakx, naky))
    call complete_alloc("fieldeqp")

    call timeadv (phi, apar, bpar, phinew, aparnew, bparnew, 0)
    call getfieldeq (phinew, aparnew, bparnew, fieldeq, fieldeqa, fieldeqp)

    irow = ifield + nfield*(ig+ntgrid)

    do ik=1,naky
       do it=1,nakx
          iflo = idx(f_lo, irow, ik, it)
          if (idx_local(f_lo, iflo)) then

             istart = 0              
             if (use_Phi) then
                ifin = istart + nidx
                istart = istart + 1
                am(istart:ifin:nfield,iflo) = fieldeq(:,it,ik) 
             end if

             if (use_Apar) then
                ifin = istart + nidx
                istart = istart + 1
                am(istart:ifin:nfield,iflo) = fieldeqa(:,it,ik)
             end if
             
             if (use_Bpar) then
                ifin = istart + nidx
                istart = istart + 1
                am(istart:ifin:nfield,iflo) = fieldeqp(:,it,ik)
             end if
          end if
       end do
    end do

    call mem_dealloc("fieldeq",tcomplex,3, 2*ntgrid+1, nakx, naky)
    call mem_dealloc("fieldeqa",tcomplex,3, 2*ntgrid+1, nakx, naky)
    call mem_dealloc("fieldeqp",tcomplex,3, 2*ntgrid+1, nakx, naky)
    deallocate (fieldeq, fieldeqa, fieldeqp)
  end subroutine init_response_row

  subroutine init_inverse_matrix (am)

    use kgrids, only: aky, akx, nakx, naky
    use theta_grid, only: ntgrid
    use fields_arrays, only: aminv
    use agk_mem, only: attempt_alloc,complete_alloc,tcomplex,mem_dealloc
    use agk_layouts, only: f_lo, idx, idx_local, proc_id
    use agk_layouts, only: if_idx, ik_idx, it_idx
    use mp, only: nproc
    implicit none
    complex, dimension(:,f_lo%llim_proc:), intent (in out) :: am
    complex, dimension(:,:,:), allocatable :: lhscol, rhsrow
    complex :: fac
    integer :: i, j, ik, it, ilo, jlo, irow

    call attempt_alloc("lhscol",tcomplex,3,nidx,nakx,naky)
    allocate (lhscol (nidx,nakx,naky))
    call complete_alloc("lhscol")

    call attempt_alloc("rhsrow",tcomplex,3,nidx,nakx,naky)
    allocate (rhsrow (nidx,nakx,naky))
    call complete_alloc("rhsrow")
       
! BD Double-check last argument
! TT: I replaced ulim_alloc - ulim_proc -> ulim_alloc - llim_proc
    call attempt_alloc("aminv", tcomplex, 2, nidx, f_lo%ulim_alloc-f_lo%llim_proc+1)
    allocate (aminv(nidx,f_lo%llim_proc:f_lo%ulim_alloc))
    call complete_alloc("aminv")

    aminv = 0.0
    do ilo = f_lo%llim_proc, f_lo%ulim_proc
       aminv(if_idx(f_lo, ilo), ilo) = 1.0
    end do
!
! Gauss-Jordan elimination: surely could be improved!
! Inverting naky*nakx arrays of size nidx*nidx in parallel
!
    do i=1,nidx
! TT>
       lhscol = 0.0; rhsrow = 0.0
! <TT
       do ik=1,naky
          do it=1,nakx
             ilo = idx(f_lo, i, ik, it)
             if (idx_local(f_lo, ilo)) then
                lhscol(:,it,ik) = am   (:,ilo)
                rhsrow(:,it,ik) = aminv(:,ilo)
             end if
! TT>
!!$             call broadcast (lhscol(:,it,ik), proc_id(f_lo, ilo))
!!$             call broadcast (rhsrow(:,it,ik), proc_id(f_lo, ilo))
! <TT
          end do
       end do
! TT>
       if (nproc > 1) then
          call sum_up_all (lhscol)
          call sum_up_all (rhsrow)
       end if
! <TT

       do jlo = f_lo%llim_proc, f_lo%ulim_proc
          irow = if_idx(f_lo,jlo)
          ik = ik_idx(f_lo,jlo)
          it = it_idx(f_lo,jlo)

          if (aky(ik) /= 0.0 .or. akx(it) /= 0.0) then
             fac = am(i,jlo)/lhscol(i,it,ik)
             am(i,jlo) = am(i,jlo)/lhscol(i,it,ik)
             am(:i-1,jlo) = am(:i-1,jlo) - lhscol(:i-1,it,ik)*fac
             am(i+1:,jlo) = am(i+1:,jlo) - lhscol(i+1:,it,ik)*fac
             
             if (irow == i) then
                aminv(:,jlo) = aminv(:,jlo)/lhscol(i,it,ik)
             else
                aminv(:,jlo) = aminv(:,jlo) &
                     - rhsrow(:,it,ik)*lhscol(irow,it,ik)/lhscol(i,it,ik)
             end if
          else
             aminv(:,jlo) = 0.
          end if
       end do
    end do

    call mem_dealloc("lhscol",tcomplex,3,nidx,nakx,naky)
    call mem_dealloc("rhsrow",tcomplex,3,nidx,nakx,naky)
    deallocate (lhscol, rhsrow)

  end subroutine init_inverse_matrix

  ! TT: We could put this subroutine in mp.fpp and use it for le_grids etc?
  subroutine sum_up_all (field)

    use agk_mem, only: attempt_alloc, complete_alloc, mem_dealloc, tcomplex
    use mp, only: sum_allreduce

    complex, dimension (:,:,:), intent (inout) :: field
    integer :: i, ik, it, ig, n1, n2, n3
    complex, dimension (:), allocatable :: work

    n1 = size (field,1)
    n2 = size (field,2)
    n3 = size (field,3)
    call attempt_alloc ("work", tcomplex, 1, n1*n2*n3)
    allocate ( work(n1*n2*n3) ) ; work = 0.
    call complete_alloc ("work")
    i = 0
    do ik=1, n3
       do it=1, n2
          do ig=1, n1
             i = i + 1
             work(i) = field(ig,it,ik)
          end do
       end do
    end do
    call sum_allreduce (work)
    i = 0
    do ik=1, n3
       do it=1, n2
          do ig=1, n1
             i = i + 1
             field(ig,it,ik) = work(i)
          end do
       end do
    end do
    call mem_dealloc ("work", tcomplex, 1, n1*n2*n3)
    deallocate (work)

  end subroutine sum_up_all

end module fields_implicit
